import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from dsp_tools.utils.rdflib_constants import SubjectObjectTypeAlias


def get_temp_directory() -> TemporaryDirectory[str]:
    ttl_dir = (Path.home() / ".dsp-tools" / "validate-data").absolute()
    ttl_dir.mkdir(exist_ok=True, parents=True)
    t_dir = TemporaryDirectory(dir=ttl_dir)
    return t_dir


def clean_up_temp_directory(temp_dir: TemporaryDirectory[str], save_graphs_dir: Path | None) -> None:
    if save_graphs_dir:
        tmp_folder = Path(temp_dir.name)
        save_graphs_dir.mkdir(parents=True, exist_ok=True)
        for tmp_filepath in tmp_folder.glob("*"):
            if tmp_filepath.is_file():
                dest_filepath = save_graphs_dir / tmp_filepath.relative_to(tmp_folder)
                shutil.copy2(tmp_filepath, dest_filepath)
    temp_dir.cleanup()


def reformat_any_iri(iri: SubjectObjectTypeAlias | str) -> str:
    """
    Reformats any kind of IRI, if it starts with data then it treats is like a data IRI.
    Otherwise, like an ontology IRI.

    Args:
        iri: Input IRI

    Returns:
        reformatted string
    """
    iri_str = str(iri)
    if iri_str.startswith("http://data/"):
        return reformat_data_iri(iri_str)
    if iri_str.startswith("http://rdfh.ch/"):
        return iri_str
    return reformat_onto_iri(iri_str)


def reformat_onto_iri(iri: SubjectObjectTypeAlias | str) -> str:
    """Takes a rdflib Node and returns a prefixed IRI in string form."""
    iri_str = str(iri)
    if "http://www.w3.org/2000/01/rdf-schema#" in iri_str:
        return f"rdfs:{iri_str.split('#')[-1]}"
    onto = iri_str.split("/")[-2]
    ending = iri_str.split("#")[-1]
    if onto == "knora-api":
        return ending
    return f"{onto}:{ending}"


def reformat_data_iri(iri: SubjectObjectTypeAlias | str) -> str:
    """Takes a rdflib Node with in the data namespace and returns only the suffix."""
    return str(iri).replace("http://data/", "", 1)



import importlib.resources
import subprocess
from pathlib import Path

import yaml
from loguru import logger
from rdflib import SH
from rdflib import Graph

from dsp_tools.commands.validate_data.models.api_responses import SHACLValidationReport
from dsp_tools.commands.validate_data.models.validation import ValidationFilePaths
from dsp_tools.error.exceptions import InternalError
from dsp_tools.error.exceptions import ShaclValidationCliError


class ShaclCliValidator:
    def validate(self, file_paths: ValidationFilePaths) -> SHACLValidationReport:
        try:
            self._run_validate_cli(file_paths)
        except subprocess.CalledProcessError as e:
            logger.error(e)
            msg = (
                "Data validation requires Docker. Is your Docker Desktop Application open? "
                "If it is, please contact the DSP-TOOLS development team with your log file."
            )
            logger.error(msg)
            raise ShaclValidationCliError(msg) from None
        return self._parse_validation_result(file_paths.directory / file_paths.report_file)

    def _run_validate_cli(self, file_paths: ValidationFilePaths) -> None:
        if not (file_paths.directory / file_paths.shacl_file).exists():
            raise InternalError(f"SHACL file not found: {file_paths.shacl_file}")
        if not (file_paths.directory / file_paths.data_file).exists():
            raise InternalError(f"Data file not found: {file_paths.data_file}")

        # Get relative paths within the container
        shacl_path = f"/data/{file_paths.shacl_file}"
        data_path = f"/data/{file_paths.data_file}"
        report_path = f"/data/{file_paths.report_file}"

        docker_file = importlib.resources.files("dsp_tools").joinpath("resources/validate_data/shacl-cli-image.yml")
        docker_spec = yaml.safe_load(docker_file.read_bytes())
        docker_image = docker_spec["image"]

        d_cmd = (
            f"docker run --rm -v {file_paths.directory.absolute()}:/data {docker_image} "
            f"validate --shacl {shacl_path} --data {data_path} --report {report_path}"
        )
        logger.debug(f"Running SHACL validation: {d_cmd}")
        result = subprocess.run(
            d_cmd.split(),
            capture_output=True,
            text=True,
            check=True,
        )
        if result.stdout:
            logger.debug(f"Validation output: {result.stdout}")
        if result.stderr:
            logger.error(f"Validation output error: {result.stderr}")

    def _parse_validation_result(self, filepath: Path) -> SHACLValidationReport:
        if not filepath.exists():
            raise InternalError(f"SHACL file not found: {filepath}")
        logger.debug(f"Parse validation response: {filepath}.")
        graph = Graph()
        graph.parse(filepath)
        conforms = bool(next(graph.objects(None, SH.conforms)))
        return SHACLValidationReport(conforms=conforms, validation_graph=graph)



import shutil
import subprocess
import unittest
from pathlib import Path

import pytest


class TestCliDepsResources(unittest.TestCase):
    """
    Test if
     - the CLI entry point works for the end user
        - Important: The setup for these tests must install DSP-TOOLS from the wheel.
     - all dependencies have been installed on the end user's machine
        - The entry point imports all modules, so a not-shipped dependency would cause an import error.
     - the resource files are accessible on the end user's machine
        - The CLI call from another working directory tests if the files in 'src/dsp_tools/resources' are accessible.
    """

    test_project_systematic_file = Path("testdata/json-project/test-project-systematic.json")
    test_project_minimal_file = Path("testdata/json-project/test-project-minimal.json")
    test_data_minimal_file = Path("testdata/xml-data/test-data-minimal.xml")
    cwd = Path("cwd")
    testdata_tmp = Path("testdata/tmp")

    @classmethod
    def setUpClass(cls) -> None:
        """Is executed once before the methods of this class are run"""
        cls.testdata_tmp.mkdir(exist_ok=True)
        cls.cwd.mkdir(exist_ok=True)

    @classmethod
    def tearDownClass(cls) -> None:
        """Is executed after the methods of this class have all run through"""
        shutil.rmtree(cls.testdata_tmp)
        shutil.rmtree(cls.cwd)
        for f in Path().glob("id2iri_*.json"):
            f.unlink()

    def test_validate_lists_section_with_schema(self) -> None:
        """Test if the resource file 'src/dsp_tools/resources/schema/lists-only.json' can be accessed."""
        args = ["create", "--lists-only", "--validate-only", str(self.test_project_systematic_file.absolute())]
        self._make_cli_call(args)

    def test_excel_to_json_resources(self) -> None:
        """
        Test if the resource file 'src/dsp_tools/resources/schema/resources-only.json' can be accessed.
        The output is not tested here, this is done in the unit tests.
        """
        excel_file = Path("testdata/excel2json/old_excel2json_files/test-name (test_label)/resources.xlsx")
        out_file = self.testdata_tmp / "_out_resources.json"
        self._make_cli_call(["excel2resources", str(excel_file.absolute()), str(out_file.absolute())])
        out_file.unlink()

    def test_excel_to_json_properties(self) -> None:
        """
        Test if the resource file 'src/dsp_tools/resources/schema/properties-only.json' can be accessed.
        The output is not tested here, this is done in the unit tests.
        """
        excel_file = Path("testdata/excel2json/old_excel2json_files/test-name (test_label)/properties.xlsx")
        out_file = self.testdata_tmp / "_out_properties.json"
        self._make_cli_call(["excel2properties", str(excel_file.absolute()), str(out_file.absolute())])
        out_file.unlink()

    def test_validate_project(self) -> None:
        """Test if the resource file 'src/dsp_tools/resources/schema/project.json' can be accessed."""
        self._make_cli_call(["create", "--validate-only", str(self.test_project_minimal_file.absolute())])

    def test_xml_upload(self) -> None:
        """Test if the resource file 'src/dsp_tools/resources/schema/data.xsd' can be accessed."""
        self._make_cli_call(["xmlupload", "--validate-only", str(self.test_data_minimal_file.absolute())])

    def _make_cli_call(self, args: list[str]) -> None:
        """
        Execute a CLI call to dsp-tools, capture its stdout and stderr,
        and raise an AssertionError with stdout and stderr if it fails.

        Args:
            args: list of arguments for the dsp-tools CLI call

        Raises:
            AssertionError: detailed diagnostic message with stdout and stderr if the call fails
        """
        try:
            subprocess.run(
                ["dsp-tools", *args],  # noqa: S607 (Starting a process with a partial executable path)
                check=True,
                capture_output=True,
                cwd=self.cwd,
            )
        except subprocess.CalledProcessError as e:
            msg = (
                f"Failed CLI call:\n'dsp-tools {' '.join(args)}'\n\n"
                f"Stdout:\n{e.output.decode('utf-8')}\n\n"
                f"Stderr:\n{e.stderr.decode('utf-8')}"
            )
            raise AssertionError(msg) from None


if __name__ == "__main__":
    pytest.main([__file__])



import shutil
from importlib.resources import as_file
from importlib.resources import files
from pathlib import Path

from loguru import logger
from rdflib import RDF
from rdflib import SH
from rdflib import Graph

from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.commands.validate_data.constants import ONTOLOGIES_DATA_TTL
from dsp_tools.commands.validate_data.constants import ONTOLOGIES_REPORT_TTL
from dsp_tools.commands.validate_data.constants import ONTOLOGIES_SHACL_TTL
from dsp_tools.commands.validate_data.models.input_problems import OntologyResourceProblem
from dsp_tools.commands.validate_data.models.input_problems import OntologyValidationProblem
from dsp_tools.commands.validate_data.models.validation import ValidationFilePaths
from dsp_tools.commands.validate_data.shacl_cli_validator import ShaclCliValidator
from dsp_tools.commands.validate_data.utils import clean_up_temp_directory
from dsp_tools.commands.validate_data.utils import get_temp_directory
from dsp_tools.commands.validate_data.utils import reformat_onto_iri
from dsp_tools.error.exceptions import ShaclValidationError
from dsp_tools.utils.rdflib_constants import SubjectObjectTypeAlias

LIST_SEPARATOR = "\n    - "


def validate_ontology(
    onto_graph: Graph, shacl_validator: ShaclCliValidator, config: ValidateDataConfig
) -> OntologyValidationProblem | None:
    """
    The API accepts erroneous cardinalities in the ontology.
    To distinguish a mistake in the data from the erroneous ontology, the ontology will be validated beforehand.
    This way, we do not have to take an erroneous ontology into account when validating the data.

    Args:
        onto_graph: the graph of the project ontologies
        shacl_validator: SHACL CLI validator
        config: The configuration where to save the information to

    Returns:
        A validation report if errors were found
    """
    tmp_dir = get_temp_directory()
    tmp_path = Path(tmp_dir.name)
    save_graph_dir = config.save_graph_dir
    try:
        result = _get_ontology_validation_result(onto_graph, shacl_validator, tmp_path)
        return result
    except Exception as e:  # noqa: BLE001
        logger.exception(e)
        save_graph_dir = tmp_path.parent / "validation-graphs"
        msg = (
            f"An error occurred during the ontology validation. "
            f"Please contact the dsp-tools development team "
            f"with your log files and the files in the directory: {save_graph_dir}"
        )
        raise ShaclValidationError(msg) from None
    finally:
        clean_up_temp_directory(tmp_dir, save_graph_dir)


def _get_ontology_validation_result(
    onto_graph: Graph, shacl_validator: ShaclCliValidator, tmp_path: Path
) -> OntologyValidationProblem | None:
    with as_file(files("dsp_tools").joinpath("resources/validate_data/validate-ontology.ttl")) as shacl_file_path:
        shacl_file = Path(shacl_file_path)
        shutil.copy(shacl_file, tmp_path / ONTOLOGIES_SHACL_TTL)
    onto_graph.serialize(tmp_path / ONTOLOGIES_DATA_TTL)
    paths = ValidationFilePaths(
        directory=tmp_path,
        data_file=ONTOLOGIES_DATA_TTL,
        shacl_file=ONTOLOGIES_SHACL_TTL,
        report_file=ONTOLOGIES_REPORT_TTL,
    )
    validation_result = shacl_validator.validate(paths)
    if validation_result.conforms:
        return None
    return OntologyValidationProblem(_reformat_ontology_validation_result(validation_result.validation_graph))


def _reformat_ontology_validation_result(validation_result: Graph) -> list[OntologyResourceProblem]:
    bns = validation_result.subjects(RDF.type, SH.ValidationResult)
    return [_get_one_problem(validation_result, bn) for bn in bns]


def _get_one_problem(val_g: Graph, result_bn: SubjectObjectTypeAlias) -> OntologyResourceProblem:
    iri = next(val_g.objects(result_bn, SH.focusNode))
    iri_str = reformat_onto_iri(iri)
    msg = str(next(val_g.objects(result_bn, SH.resultMessage)))
    splt = [x.strip() for x in msg.split("\n")]
    return OntologyResourceProblem(iri_str, " ".join(splt))


def get_msg_str_ontology_validation_violation(onto_violations: OntologyValidationProblem) -> str:
    probs = sorted(onto_violations.problems, key=lambda x: x.res_iri)

    def get_resource_msg(res: OntologyResourceProblem) -> str:
        return f"Resource Class: {res.res_iri} | Problem: {res.msg}"

    problems = [get_resource_msg(x) for x in probs]
    return (
        "The ontology structure contains errors that prevent the validation of the data.\n"
        "Please correct the following errors and re-upload the corrected ontology.\n"
        f"Once those two steps are done, the command `validate-data` will find any problems in the data.\n"
        f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(problems)}"
    )



# mypy: disable-error-code="method-assign,no-untyped-def"

import warnings

import numpy as np
import pandas as pd
import pytest
import regex

from dsp_tools.error.xmllib_warnings import XmllibInputInfo
from dsp_tools.error.xmllib_warnings import XmllibInputWarning
from dsp_tools.xmllib.internal.checkers import check_and_inform_about_angular_brackets
from dsp_tools.xmllib.internal.checkers import check_and_warn_potentially_empty_string


@pytest.mark.parametrize("in_val", ["string", 2, False])
def test_check_and_warn_potentially_empty_string_good(in_val):
    with warnings.catch_warnings(record=True) as caught_warnings:
        check_and_warn_potentially_empty_string(value=in_val, res_id="res", expected="input")
    assert len(caught_warnings) == 0


@pytest.mark.parametrize("in_val", ["  ", pd.NA, None])
def test_check_and_warn_potentially_empty_string_empty(in_val):
    expected = regex.escape(rf"Your input '{in_val}' is empty. Please enter a valid input.")
    with pytest.warns(XmllibInputWarning, match=expected):
        check_and_warn_potentially_empty_string(value=in_val, res_id="res", expected="input")


@pytest.mark.parametrize(
    ("in_val", "type_for_message"), [(str(pd.NA), "pd.NA"), (str(np.nan), "np.nan"), (str(None), "None")]
)
def test_check_and_warn_potentially_empty_string_potentially_empty(in_val, type_for_message):
    expected = regex.escape(
        rf"Your input '{in_val}' is a string but may be the result of `str({type_for_message})`. "
        r"Please verify that the input is as expected."
    )
    with pytest.warns(XmllibInputInfo, match=expected):
        check_and_warn_potentially_empty_string(value=in_val, res_id="res", expected="input")


@pytest.mark.parametrize(
    "allowed_values",
    [
        "(<2cm) (>10cm)",
        "text < text/>",
        "text < text> & text",
        "text <text text > text",
        'text < text text="text"> text',
        'text <text text="text" > text',
    ],
)
def test_check_and_inform_about_angular_brackets_good(allowed_values, recwarn: pytest.WarningsRecorder):
    check_and_inform_about_angular_brackets(allowed_values, "id", "prp")
    assert len(recwarn) == 0


@pytest.mark.parametrize("in_val", ["text <tag> text", "text <p>paragraph</p> text"])
def test_check_and_inform_about_angular_brackets_informs(in_val):
    expected = regex.escape(
        rf"Your input '{in_val}' contains angular brackets. "
        r"Since this is a simpletext, please note that these will not be recognised as formatting "
        r"in the text field, but will be displayed as-is."
    )
    with pytest.warns(XmllibInputInfo, match=expected):
        check_and_inform_about_angular_brackets(in_val, "id", "prp")



from collections import defaultdict

from loguru import logger

from dsp_tools.commands.validate_data.models.input_problems import DuplicateFileWarning
from dsp_tools.commands.validate_data.models.input_problems import InputProblem
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import Severity
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource


def check_for_duplicate_files(parsed_resources: list[ParsedResource]) -> DuplicateFileWarning | None:
    """
    Too many duplicate filepaths in the data may cause the SHACL validator to crash.
    If one file is referenced n times, this produces n * (n-1) validation errors.
    Therefore, this programmatic pre-validation prevents crashes during later validation steps.

    Args:
        parsed_resources: Resources to check

    Returns:
        Results for the user and decisions how the program should continue
    """
    count_dict = _get_filepaths_with_more_than_one_usage(parsed_resources)
    if not count_dict:
        return None
    input_problems = _create_input_problems(count_dict)
    return DuplicateFileWarning(input_problems)


def _get_filepaths_with_more_than_one_usage(parsed_resources: list[ParsedResource]) -> dict[str, int]:
    count_dict: dict[str, int] = defaultdict(int)
    for res in parsed_resources:
        if res.file_value and res.file_value.value:
            count_dict[res.file_value.value] += 1
    return {f_path: count for f_path, count in count_dict.items() if count > 1}


def _create_input_problems(duplicates: dict[str, int]) -> list[InputProblem]:
    all_duplicates = []
    for dup_entry, usage_count in duplicates.items():
        msg = f"value used {usage_count} times"
        logger.warning(f"File '{dup_entry}' {msg}")
        all_duplicates.append(
            InputProblem(
                problem_type=ProblemType.FILE_DUPLICATE,
                res_id=None,
                res_type=None,
                prop_name="bitstream / iiif-uri",
                severity=Severity.WARNING,
                message=msg,
                input_value=dup_entry,
            )
        )
    return all_duplicates



# mypy: disable-error-code="method-assign,no-untyped-def"
import warnings

import numpy as np
import pandas as pd
import pytest

from dsp_tools.error.xmllib_warnings import XmllibInputInfo
from dsp_tools.error.xmllib_warnings import XmllibInputWarning
from dsp_tools.xmllib.internal.input_converters import check_and_fix_is_non_empty_string
from dsp_tools.xmllib.internal.input_converters import numeric_entities


@pytest.mark.parametrize(
    ("original", "expected"),
    [
        ("a &nbsp; a", "a &#160; a"),
        ("a &#160; a", "a &#160; a"),
        ("a &#x22; a", "a &#x22; a"),
        ("a &quot; &amp; &apos; &lt; &gt; a", "a &quot; &amp; &apos; &lt; &gt; a"),
        ("aäö&;", "aäö&;"),
    ],
)
def test_numeric_entities(original: str, expected: str) -> None:
    assert numeric_entities(original) == expected


@pytest.mark.parametrize(
    ("original", "expected"),
    [("Text", "Text"), (1, "1")],
)
def test_check_and_fix_is_non_empty_string_good(original: str, expected: str):
    with warnings.catch_warnings(record=True) as caught_warnings:
        result = check_and_fix_is_non_empty_string(original)
    assert len(caught_warnings) == 0
    assert result == expected


@pytest.mark.parametrize("original", [str(pd.NA), str(None)])
def test_check_and_fix_is_non_empty_string_info(original: str):
    with pytest.warns(XmllibInputInfo):
        result = check_and_fix_is_non_empty_string(original)
    assert result == original


@pytest.mark.parametrize("original", [pd.NA, np.nan, None, " ", ""])
def test_check_and_fix_is_non_empty_string_warns(original: str):
    with pytest.warns(XmllibInputWarning):
        result = check_and_fix_is_non_empty_string(original)
    assert result == ""



from uuid import uuid4

import pytest

from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.processed.res import ProcessedResource
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedLink
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedRichtext
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedSimpleText
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedValue
from dsp_tools.commands.xmlupload.stash.create_info_for_graph import _process_one_resource
from dsp_tools.commands.xmlupload.stash.create_info_for_graph import create_info_for_graph_from_processed_resources


@pytest.fixture
def link_value() -> ProcessedValue:
    return ProcessedLink("res_id_target", "prop", None, None, str(uuid4()))


@pytest.fixture
def text_value_with_link() -> ProcessedValue:
    return ProcessedRichtext(
        FormattedTextValue(
            'This text contains a link: <a class="salsah-link" href="IRI:res_id_target:IRI">target resource</a>'
        ),
        "prop",
        None,
        None,
        {"res_id_target"},
        str(uuid4()),
    )


@pytest.fixture
def text_value_no_link() -> ProcessedValue:
    return ProcessedRichtext(FormattedTextValue("Text"), "prop", None, None, set(), str(uuid4()))


@pytest.fixture
def resource_with_link_and_text(link_value: ProcessedValue, text_value_with_link: ProcessedValue) -> ProcessedResource:
    return ProcessedResource("res_id", "res_type", "lbl", None, [link_value, text_value_with_link])


@pytest.fixture
def resource_with_link(link_value: ProcessedValue) -> ProcessedResource:
    return ProcessedResource("res_id", "res_type", "lbl", None, [link_value])


@pytest.fixture
def resource_with_text(text_value_with_link: ProcessedValue) -> ProcessedResource:
    return ProcessedResource("res_id", "res_type", "lbl", None, [text_value_with_link])


@pytest.fixture
def resource_without_links() -> ProcessedResource:
    return ProcessedResource(
        "res_id_target",
        "res_type",
        "lbl",
        None,
        [ProcessedSimpleText("val", "prop", None, None)],
    )


def test_create_info_for_graph_from_data(
    resource_without_links: ProcessedResource, resource_with_link_and_text: ProcessedResource
) -> None:
    data = [resource_with_link_and_text, resource_without_links]
    result = create_info_for_graph_from_processed_resources(data)
    assert set(result.all_resource_ids) == {"res_id_target", "res_id"}
    assert len(result.link_values) == 1
    assert len(result.standoff_links) == 1


def test_process_one_resource_no_links(resource_without_links: ProcessedResource) -> None:
    links, standoff = _process_one_resource(resource_without_links)
    assert not links
    assert not standoff


def test_process_one_resource_stand_off(
    resource_with_text: ProcessedResource, text_value_with_link: ProcessedRichtext
) -> None:
    links, standoff_list = _process_one_resource(resource_with_text)
    assert not links
    assert len(standoff_list) == 1
    standoff = standoff_list.pop(0)
    assert standoff.source_id == "res_id"
    assert standoff.target_ids == {"res_id_target"}
    assert standoff.link_uuid == text_value_with_link.value_uuid


def test_process_one_resource_link_values(resource_with_link: ProcessedResource, link_value: ProcessedLink) -> None:
    link_list, standoff = _process_one_resource(resource_with_link)
    assert not standoff
    assert len(link_list) == 1
    link = link_list.pop(0)
    assert link.source_id == "res_id"
    assert link.target_id == "res_id_target"
    assert link.link_uuid == link_value.value_uuid


def test_process_one_resource_both_links(
    resource_with_link_and_text: ProcessedResource,
    link_value: ProcessedLink,
    text_value_with_link: ProcessedRichtext,
) -> None:
    link_list, standoff_list = _process_one_resource(resource_with_link_and_text)
    assert len(link_list) == 1
    link = link_list.pop(0)
    assert link.source_id == "res_id"
    assert link.target_id == "res_id_target"
    assert link.link_uuid == link_value.value_uuid

    assert len(standoff_list) == 1
    standoff = standoff_list.pop(0)
    assert standoff.source_id == "res_id"
    assert standoff.target_ids == {"res_id_target"}
    assert standoff.link_uuid == text_value_with_link.value_uuid


def test_process_one_resource_with_iris() -> None:
    resource_iri = "http://rdfh.ch/4123/wFnAqlqXQC69SBO0z27-4A"
    richtext = ProcessedRichtext(
        FormattedTextValue(
            f'This text contains a link: <a class="salsah-link" href="{resource_iri}">target resource</a>'
        ),
        "prop",
        None,
        None,
        {resource_iri},
        str(uuid4()),
    )
    link_value = ProcessedLink(resource_iri, "prop", None, None, str(uuid4()))
    resource = ProcessedResource("res_id", "res_type", "lbl", None, [link_value, richtext])
    link_list, standoff_list = _process_one_resource(resource)
    assert not link_list
    assert not standoff_list


def test_process_one_resource_with_mixed_iri_and_id() -> None:
    resource_iri = "http://rdfh.ch/4123/wFnAqlqXQC69SBO0z27-4A"
    richtext = ProcessedRichtext(
        FormattedTextValue(
            f'This text contains a link: <a class="salsah-link" href="{resource_iri}">target resource</a>'
            f'and a link with an id: <a class="salsah-link" href="IRI:res_id_target:IRI">target resource</a>'
        ),
        "prop",
        None,
        None,
        {resource_iri, "res_id_target"},
        str(uuid4()),
    )
    resource = ProcessedResource("res_id", "res_type", "lbl", None, [richtext])
    link_list, standoff_list = _process_one_resource(resource)
    assert not link_list
    assert len(standoff_list) == 1
    standoff = standoff_list.pop(0)
    assert standoff.source_id == "res_id"
    assert standoff.target_ids == {"res_id_target"}
    assert standoff.link_uuid == richtext.value_uuid


if __name__ == "__main__":
    pytest.main([__file__])



import shutil
from collections.abc import Iterator
from pathlib import Path

import pytest
import regex

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.ingest_xmlupload.create_resources.upload_xml import ingest_xmlupload
from dsp_tools.error.exceptions import InputError


@pytest.fixture(autouse=True)
def _retrieve_mapping_file() -> Iterator[None]:
    """Put the mapping file into the cwd."""
    mapping_file = Path("testdata/dsp-ingest-data/mapping-00A2.csv")
    shutil.copy(mapping_file, ".")
    yield
    Path(mapping_file.name).unlink()


@pytest.fixture
def creds() -> ServerCredentials:
    return ServerCredentials(
        user="root@example.com",
        password="test",
        server="http://0.0.0.0:3333",
        dsp_ingest_url="http://0.0.0.0:3340",
    )


def test_ingest_xmlupload(creds: ServerCredentials) -> None:
    expected_msg = regex.escape(
        "The upload cannot continue as there are problems with the multimedia files referenced in the XML.\n"
        "    The data XML file does not reference the following multimedia files "
        "which were previously uploaded through dsp-ingest:\n"
        "    - extra.jpg\n"
        "    The data XML file contains references to the following multimedia files "
        "which were not previously uploaded through dsp-ingest:\n"
        "    - Resource ID: 'GoodGirlImage' | Filepath: 'images/GoodGirl.jpg'"
    )
    with pytest.raises(InputError, match=expected_msg):
        ingest_xmlupload(xml_file=Path("testdata/dsp-ingest-data/dsp-ingest.xml"), creds=creds)


def test_ingest_xmlupload_no_mapping(creds: ServerCredentials) -> None:
    expected_msg = regex.escape("No mapping CSV file was found at mapping-00A5.csv.")
    with pytest.raises(InputError, match=expected_msg):
        ingest_xmlupload(xml_file=Path("testdata/dsp-ingest-data/dsp_ingest_no_mapping.xml"), creds=creds)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from enum import StrEnum
from enum import auto

import pandas as pd
import regex

from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs


@dataclass
class ValidateDataResult:
    no_problems: bool
    problems: None | UnknownClassesInData | OntologyValidationProblem | SortedProblems
    report_graphs: None | ValidationReportGraphs


@dataclass
class OntologyValidationProblem:
    problems: list[OntologyResourceProblem]


@dataclass
class OntologyResourceProblem:
    res_iri: str
    msg: str


@dataclass
class UnknownClassesInData:
    unknown_classes: set[str]
    defined_classes: set[str]


@dataclass
class DuplicateFileWarning:
    problems: list[InputProblem]


@dataclass
class AllProblems:
    problems: list[InputProblem]
    unexpected_results: list[UnexpectedComponent]


@dataclass
class SortedProblems:
    unique_violations: list[InputProblem]
    user_warnings: list[InputProblem]
    user_info: list[InputProblem]
    unexpected_shacl_validation_components: list[str]


@dataclass
class MessageComponents:
    message_header: str
    message_body: str | None
    message_df: pd.DataFrame | None


@dataclass
class UserPrintMessages:
    violations: MessageComponents | None
    warnings: MessageComponents | None
    infos: MessageComponents | None
    unexpected_violations: MessageComponents | None


@dataclass
class InputProblem:
    problem_type: ProblemType
    res_id: str | None
    res_type: str | None
    prop_name: str
    severity: Severity
    message: str | None = None
    input_value: str | None = None
    input_type: str | None = None
    expected: str | None = None

    def __post_init__(self) -> None:
        self.message = self._clean_input(self.message)
        self.expected = self._clean_input(self.expected)

    def _clean_input(self, to_clean: str | None) -> str | None:
        if not to_clean:
            return None
        if not regex.search(r"\S+", to_clean):
            return None
        str_split = [content for x in to_clean.split(" ") if (content := x.strip())]
        return " ".join(str_split)


class Severity(Enum):
    VIOLATION = auto()
    WARNING = auto()
    INFO = auto()


class ProblemType(StrEnum):
    GENERIC = "generic"
    FILE_VALUE_MISSING = "file problem"
    FILE_DUPLICATE = "Duplicate Filepath / IIIF-URI"
    FILE_VALUE_PROHIBITED = "A file was added to the resource. This resource type must not have a file."
    MAX_CARD = "Maximum Cardinality Violation"
    MIN_CARD = "Minimum Cardinality Violation"
    NON_EXISTING_CARD = "The resource class does not have a cardinality for this property."
    VALUE_TYPE_MISMATCH = "Value Type Mismatch"
    INPUT_REGEX = "Wrong Format of Input"
    LINK_TARGET_TYPE_MISMATCH = "Linked Resource Type Mismatch"
    INEXISTENT_LINKED_RESOURCE = "Linked Resource does not exist"
    DUPLICATE_VALUE = "Your input is duplicated"



from __future__ import annotations

from lxml import etree

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.xmllib import escape_reserved_xml_characters
from dsp_tools.xmllib.internal.input_converters import numeric_entities


def parse_richtext_as_xml(
    input_str: str, resource_id: str | None = None, prop_name: str | None = None
) -> etree._Element | MessageInfo:
    """
    Parses an input string as XML. If it contains a syntax error a MessageInfo is returned.
    Else it returns the XML as etree

    Args:
        input_str: Richtext string
        resource_id: ID of the resource for improved error message
        prop_name: Property name for improved error message

    Returns:
        Parsed string or information for the user message.
    """
    escaped_text = escape_reserved_xml_characters(input_str)
    num_ent = numeric_entities(escaped_text)
    pseudo_xml = f"<ignore-this>{num_ent}</ignore-this>"
    try:
        return etree.fromstring(pseudo_xml)
    except etree.XMLSyntaxError as err:
        msg_str = (
            f"The entered richtext value could not be converted to a valid XML.\n"
            f"Original error message: {err.msg}\n"
            f"Potential line/column numbers are relative to this text: {pseudo_xml}"
        )
        return MessageInfo(resource_id=resource_id, prop_name=prop_name, message=msg_str)



from __future__ import annotations

from loguru import logger
from lxml import etree

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.xmlupload.models.lookup_models import XmlReferenceLookups
from dsp_tools.commands.xmlupload.models.processed.res import ProcessedResource
from dsp_tools.commands.xmlupload.models.upload_clients import UploadClients
from dsp_tools.commands.xmlupload.stash.analyse_circular_reference_graph import generate_upload_order
from dsp_tools.commands.xmlupload.stash.create_info_for_graph import create_info_for_graph_from_processed_resources
from dsp_tools.commands.xmlupload.stash.stash_circular_references import stash_circular_references
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.error.exceptions import BaseError
from dsp_tools.error.exceptions import InputError
from dsp_tools.legacy_models.projectContext import ProjectContext
from dsp_tools.utils.xml_parsing.get_lookups import get_authorship_lookup
from dsp_tools.utils.xml_parsing.get_lookups import get_permissions_lookup
from dsp_tools.utils.xml_parsing.get_parsed_resources import get_parsed_resources
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource

LIST_SEPARATOR = "\n-    "


def get_parsed_resources_and_mappers(
    root: etree._Element, clients: UploadClients
) -> tuple[list[ParsedResource], XmlReferenceLookups]:
    logger.debug("Get ParsedResource and XML-Lookups from root")
    parsed_resources = get_parsed_resources(root, clients.legal_info_client.server)
    processed_lookups = _get_xml_reference_lookups(root=root, clients=clients)
    return parsed_resources, processed_lookups


def _get_xml_reference_lookups(root: etree._Element, clients: UploadClients) -> XmlReferenceLookups:
    proj_context = _get_project_context_from_server(
        connection=clients.project_client.con, shortcode=root.attrib["shortcode"]
    )
    permissions_lookup = get_permissions_lookup(root, proj_context)
    authorship_lookup = get_authorship_lookup(root)
    listnode_lookup = clients.list_client.get_list_node_id_to_iri_lookup()
    return XmlReferenceLookups(
        permissions=permissions_lookup,
        listnodes=listnode_lookup,
        authorships=authorship_lookup,
    )


def _get_project_context_from_server(connection: Connection, shortcode: str) -> ProjectContext:
    try:
        proj_context = ProjectContext(con=connection, shortcode=shortcode)
    except BaseError:
        logger.exception("Unable to retrieve project context from DSP server")
        raise InputError("Unable to retrieve project context from DSP server") from None
    return proj_context


def get_stash_and_upload_order(
    resources: list[ProcessedResource],
) -> tuple[list[ProcessedResource], Stash | None]:
    logger.debug("Get stash and upload order")
    info_for_graph = create_info_for_graph_from_processed_resources(resources)
    stash_lookup, upload_order = generate_upload_order(info_for_graph)
    sorting_lookup = {res.res_id: res for res in resources}
    sorted_resources = [sorting_lookup[res_id] for res_id in upload_order]
    stash = stash_circular_references(sorted_resources, stash_lookup)
    return sorted_resources, stash



import unittest
from typing import Union

import numpy as np
import pandas as pd
import pytest

from dsp_tools import excel2xml

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


class TestPrepareValue(unittest.TestCase):
    def test_prepare_value(self) -> None:
        identical_values = ["Test", "Test", "Test"]
        different_values: list[Union[str, int, float]] = [1, 1.0, "1", "1.0", " 1 "]
        values_with_nas: list[Union[str, int, float]] = ["test", "", 1, np.nan, pd.NA, 0]  # type: ignore[list-item]

        for val in different_values:
            values_output = excel2xml.prepare_value(val)
            self.assertEqual([x.value for x in values_output], [val])

            values_output = excel2xml.prepare_value(excel2xml.PropertyElement(val))
            self.assertEqual([x.value for x in values_output], [val])

        values_output = excel2xml.prepare_value(identical_values)
        self.assertEqual([x.value for x in values_output], identical_values)

        values_output = excel2xml.prepare_value([excel2xml.PropertyElement(x) for x in identical_values])
        self.assertEqual([x.value for x in values_output], identical_values)

        values_output = excel2xml.prepare_value(different_values)
        self.assertEqual([x.value for x in values_output], different_values)

        values_output = excel2xml.prepare_value([excel2xml.PropertyElement(x) for x in different_values])
        self.assertEqual([x.value for x in values_output], different_values)

        values_output = excel2xml.prepare_value(values_with_nas)
        self.assertEqual([x.value for x in values_output], values_with_nas)

        values_output = excel2xml.prepare_value([excel2xml.PropertyElement(x) for x in values_with_nas])
        self.assertEqual([x.value for x in values_output], values_with_nas)


if __name__ == "__main__":
    pytest.main([__file__])



from enum import Enum

from lxml import etree

from dsp_tools.xmllib.internal.constants import DASCH_SCHEMA
from dsp_tools.xmllib.internal.constants import XML_NAMESPACE_MAP


class PermissionTypes(Enum):
    RV = "RV"
    V = "V"
    D = "D"
    CR = "CR"


PUBLIC = {
    "UnknownUser": PermissionTypes.V,
    "KnownUser": PermissionTypes.V,
    "ProjectMember": PermissionTypes.D,
    "ProjectAdmin": PermissionTypes.CR,
}

PRIVATE = {
    "ProjectMember": PermissionTypes.D,
    "ProjectAdmin": PermissionTypes.CR,
}

LIMITED_VIEW = {
    "UnknownUser": PermissionTypes.RV,
    "KnownUser": PermissionTypes.RV,
    "ProjectMember": PermissionTypes.D,
    "ProjectAdmin": PermissionTypes.CR,
}


class XMLPermissions:
    def serialise(self, contains_old_permissions: bool, contains_new_permissions: bool = True) -> list[etree._Element]:
        new_perms = [
            self._serialise_one_permission_element("public", PUBLIC),
            self._serialise_one_permission_element("private", PRIVATE),
            self._serialise_one_permission_element("limited_view", LIMITED_VIEW),
        ]
        old_perms = [
            self._serialise_one_permission_element("open", PUBLIC),
            self._serialise_one_permission_element("restricted", PRIVATE),
            self._serialise_one_permission_element("restricted-view", LIMITED_VIEW),
        ]
        res = []
        if contains_new_permissions:
            res.extend(new_perms)
        if contains_old_permissions:
            res.extend(old_perms)
        return res

    def _serialise_one_permission_element(
        self, permission_name: str, groups: dict[str, PermissionTypes]
    ) -> etree._Element:
        ele = etree.Element(f"{DASCH_SCHEMA}permissions", attrib={"id": permission_name}, nsmap=XML_NAMESPACE_MAP)
        allows = [self._serialise_one_allow(g, name) for g, name in groups.items()]
        ele.extend(allows)
        return ele

    def _serialise_one_allow(self, group: str, tag_text: PermissionTypes) -> etree._Element:
        ele = etree.Element(f"{DASCH_SCHEMA}allow", attrib={"group": group}, nsmap=XML_NAMESPACE_MAP)
        ele.text = tag_text.value
        return ele



import json
from json import JSONDecodeError
from typing import Any
from typing import cast

import regex

from dsp_tools.commands.validate_data.mappers import FILE_TYPE_TO_PROP
from dsp_tools.commands.validate_data.models.api_responses import ListLookup
from dsp_tools.commands.validate_data.models.rdf_like_data import MigrationMetadata
from dsp_tools.commands.validate_data.models.rdf_like_data import PropertyObject
from dsp_tools.commands.validate_data.models.rdf_like_data import RdfLikeData
from dsp_tools.commands.validate_data.models.rdf_like_data import RdfLikeResource
from dsp_tools.commands.validate_data.models.rdf_like_data import RdfLikeValue
from dsp_tools.commands.validate_data.models.rdf_like_data import TripleObjectType
from dsp_tools.commands.validate_data.models.rdf_like_data import TriplePropertyType
from dsp_tools.utils.data_formats.date_util import Era
from dsp_tools.utils.data_formats.date_util import SingleDate
from dsp_tools.utils.data_formats.date_util import parse_date_string
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValueMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedValue


def get_rdf_like_data(
    resources: list[ParsedResource], authorship_lookup: dict[str, list[str]], list_node_lookup: ListLookup
) -> RdfLikeData:
    rdf_like_resources = [_get_one_resource(x, authorship_lookup, list_node_lookup) for x in resources]
    return RdfLikeData(rdf_like_resources)


def _get_one_resource(
    resource: ParsedResource, authorship_lookup: dict[str, list[str]], list_node_lookup: ListLookup
) -> RdfLikeResource:
    values = [_get_one_value(x, list_node_lookup) for x in resource.values]
    if resource.file_value:
        if file_val := _get_file_value(resource.file_value, authorship_lookup):
            values.append(file_val)
    metadata = [
        PropertyObject(TriplePropertyType.RDFS_LABEL, resource.label, TripleObjectType.STRING),
        PropertyObject(TriplePropertyType.RDF_TYPE, resource.res_type, TripleObjectType.IRI),
    ]
    metadata.extend(_get_all_stand_off_links(values))
    if resource.permissions_id is not None:
        metadata.append(
            PropertyObject(TriplePropertyType.KNORA_PERMISSIONS, resource.permissions_id, TripleObjectType.STRING)
        )
    return RdfLikeResource(
        res_id=resource.res_id,
        property_objects=metadata,
        values=values,
        migration_metadata=MigrationMetadata(),
    )


def _get_all_stand_off_links(values: list[RdfLikeValue]) -> list[PropertyObject]:
    stand_offs = []
    for val in values:
        if val.knora_type.RICHTEXT_VALUE:
            stand_offs.extend(_get_stand_off_links(val.user_facing_value))
    return stand_offs


def _get_stand_off_links(text: str | None) -> list[PropertyObject]:
    if not text:
        return []
    links = set(regex.findall(pattern='href="IRI:(.*?):IRI"', string=text))
    return [PropertyObject(TriplePropertyType.KNORA_STANDOFF_LINK, lnk, TripleObjectType.IRI) for lnk in links]


def _get_one_value(value: ParsedValue, list_node_lookup: ListLookup) -> RdfLikeValue:
    user_value = value.value
    match value.value_type:
        case KnoraValueType.DATE_VALUE:
            return _get_date_value(value)
        case KnoraValueType.INTERVAL_VALUE:
            return _get_interval_value(value)
        case KnoraValueType.LIST_VALUE:
            user_value = _get_list_value_str(user_value, list_node_lookup)
        case KnoraValueType.GEOM_VALUE:
            user_value = _get_geometry_value_str(user_value)
        case _:
            pass
    typed_val: str | None = user_value if isinstance(user_value, str) else None
    return _get_generic_value(value, typed_val)


def _get_generic_value(value: ParsedValue, user_value: str | None) -> RdfLikeValue:
    return RdfLikeValue(
        user_facing_prop=value.prop_name,
        user_facing_value=user_value,
        knora_type=value.value_type,
        value_metadata=_get_value_metadata(value),
    )


def _get_date_value(value: ParsedValue) -> RdfLikeValue:
    typed_val: str | None = value.value if isinstance(value.value, str) else None
    date_metadata = _get_value_metadata(value)
    if typed_val:
        date_metadata.extend(_get_xsd_like_dates(typed_val))
    return RdfLikeValue(
        user_facing_prop=value.prop_name,
        user_facing_value=typed_val,
        knora_type=value.value_type,
        value_metadata=date_metadata,
    )


def _get_xsd_like_dates(date_string: str) -> list[PropertyObject]:
    parsed_date = parse_date_string(date_string)
    dates = []
    if not (ce_start := _make_xsd_compatible_date(parsed_date.start, TriplePropertyType.KNORA_DATE_START)):
        return []
    dates.append(ce_start)
    if parsed_date.end:
        if ce_end := _make_xsd_compatible_date(parsed_date.end, TriplePropertyType.KNORA_DATE_END):
            dates.append(ce_end)
    return dates


def _make_xsd_compatible_date(single_date: SingleDate, prop_type: TriplePropertyType) -> PropertyObject | None:
    if single_date.era in (Era.BC, Era.BCE):
        return None
    date_str = _get_date_str(single_date)
    return PropertyObject(
        property_type=prop_type,
        object_value=date_str,
        object_type=TripleObjectType.DATE_YYYY_MM_DD,
    )


def _get_date_str(date: SingleDate) -> str:
    # SHACL cannot compare dates of varying precision, therefore we turn partial dates into full dates
    date_str = [str(date.year).zfill(4)]
    if date.month:
        date_str.append(str(date.month).zfill(2))
    else:
        date_str.append("01")
    if date.day:
        date_str.append(str(date.day).zfill(2))
    else:
        date_str.append("01")
    return "-".join(date_str)


def _get_interval_value(value: ParsedValue) -> RdfLikeValue:
    property_objects = []
    tuple_val = value.value
    if isinstance(tuple_val, tuple):
        if first := tuple_val[0]:
            property_objects.append(
                PropertyObject(
                    property_type=TriplePropertyType.KNORA_INTERVAL_START,
                    object_value=first,
                    object_type=TripleObjectType.DECIMAL,
                )
            )
        if second := tuple_val[1]:
            property_objects.append(
                PropertyObject(
                    property_type=TriplePropertyType.KNORA_INTERVAL_END,
                    object_value=second,
                    object_type=TripleObjectType.DECIMAL,
                )
            )
    property_objects.extend(_get_value_metadata(value))
    return RdfLikeValue(
        user_facing_prop=value.prop_name,
        user_facing_value=None,
        knora_type=value.value_type,
        value_metadata=property_objects,
    )


def _get_list_value_str(user_value: str | tuple[str | None, str | None] | None, list_node_lookup: ListLookup) -> str:
    in_tuple = cast(tuple[Any, Any], user_value)
    if found := list_node_lookup.lists.get(in_tuple):
        return found
    return " / ".join(x for x in in_tuple if x is not None)


def _get_geometry_value_str(user_value: str | tuple[str | None, str | None] | None) -> str | None:
    try:
        if isinstance(user_value, str):
            return json.dumps(json.loads(user_value))
        return None
    except JSONDecodeError:
        return None


def _get_value_metadata(value: ParsedValue) -> list[PropertyObject]:
    metadata = []
    if value.permissions_id is not None:
        metadata.append(
            PropertyObject(
                property_type=TriplePropertyType.KNORA_PERMISSIONS,
                object_value=value.permissions_id,
                object_type=TripleObjectType.STRING,
            )
        )
    if value.comment is not None:
        metadata.append(
            PropertyObject(
                property_type=TriplePropertyType.KNORA_COMMENT_ON_VALUE,
                object_value=value.comment,
                object_type=TripleObjectType.STRING,
            )
        )
    return metadata


def _get_file_value(file_value: ParsedFileValue, authorship_lookup: dict[str, list[str]]) -> RdfLikeValue | None:
    if not all([file_value.value, file_value.value_type]):
        return None
    file_type = cast(KnoraValueType, file_value.value_type)
    user_prop = FILE_TYPE_TO_PROP[file_type]
    return RdfLikeValue(
        user_facing_prop=user_prop,
        user_facing_value=file_value.value,
        knora_type=file_type,
        value_metadata=_get_file_metadata(file_value.metadata, authorship_lookup),
    )


def _get_file_metadata(
    metadata: ParsedFileValueMetadata, authorship_lookup: dict[str, list[str]]
) -> list[PropertyObject]:
    property_objects = []
    if metadata.license_iri is not None:
        property_objects.append(
            PropertyObject(
                property_type=TriplePropertyType.KNORA_LICENSE,
                object_value=metadata.license_iri,
                object_type=TripleObjectType.IRI,
            )
        )
    if metadata.copyright_holder is not None:
        property_objects.append(
            PropertyObject(
                property_type=TriplePropertyType.KNORA_COPYRIGHT_HOLDER,
                object_value=metadata.copyright_holder,
                object_type=TripleObjectType.STRING,
            )
        )
    if metadata.authorship_id is not None:
        if not (found_authors := authorship_lookup.get(metadata.authorship_id)):
            property_objects.append(
                PropertyObject(
                    property_type=TriplePropertyType.KNORA_AUTHORSHIP,
                    object_value="",
                    object_type=TripleObjectType.STRING,
                )
            )
        else:
            for auth in found_authors:
                property_objects.append(
                    PropertyObject(
                        property_type=TriplePropertyType.KNORA_AUTHORSHIP,
                        object_value=auth,
                        object_type=TripleObjectType.STRING,
                    )
                )
    if metadata.permissions_id is not None:
        property_objects.append(
            PropertyObject(
                property_type=TriplePropertyType.KNORA_PERMISSIONS,
                object_value=metadata.permissions_id,
                object_type=TripleObjectType.STRING,
            )
        )
    return property_objects



import warnings

import pytest
import regex
from lxml import etree

from dsp_tools.xmllib.internal.serialise_resource import _serialise_one_resource
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.dsp_base_resources import LinkResource
from dsp_tools.xmllib.models.dsp_base_resources import RegionResource
from dsp_tools.xmllib.models.internal.file_values import AuthorshipLookup
from dsp_tools.xmllib.models.licenses.other import LicenseOther
from dsp_tools.xmllib.models.licenses.recommended import LicenseRecommended
from dsp_tools.xmllib.models.res import Resource

AUTHOR_LOOKUP = AuthorshipLookup({("one", "one2"): "authorship_1"})


@pytest.fixture
def region_no_warnings() -> RegionResource:
    region = RegionResource.create_new("res_id", "label", "region_of")
    return region.add_rectangle((0.1, 0.1), (0.2, 0.2))


@pytest.fixture
def link_obj_no_warnings() -> LinkResource:
    return LinkResource.create_new("id", "lbl", ["link"]).add_comment("cmt")


class TestResource:
    def test_no_values(self) -> None:
        res = Resource.create_new("id", ":Type", "lbl")
        serialised = etree.tostring(_serialise_one_resource(res, AUTHOR_LOOKUP))
        expected = (
            b'<resource xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'label="lbl" id="id" restype=":Type"/>'
        )
        assert serialised == expected

    def test_permissions(self) -> None:
        res = Resource.create_new("id", ":Type", "lbl", permissions=Permissions.PUBLIC)
        serialised = etree.tostring(_serialise_one_resource(res, AUTHOR_LOOKUP))
        expected = (
            b'<resource xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'label="lbl" id="id" permissions="public" restype=":Type"/>'
        )
        assert serialised == expected

    def test_one_value(self) -> None:
        res = Resource.create_new("id", ":Type", "lbl").add_bool(":bool", True)
        serialised = etree.tostring(_serialise_one_resource(res, AUTHOR_LOOKUP))
        expected = (
            b'<resource xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'label="lbl" id="id" restype=":Type">'
            b'<boolean-prop name=":bool"><boolean>true</boolean></boolean-prop>'
            b"</resource>"
        )
        assert serialised == expected

    def test_serialise_no_warnings(self) -> None:
        res = Resource.create_new("id", ":Type", "lbl").add_file(
            "file.jpg", LicenseRecommended.DSP.UNKNOWN, "copy", ["one", "one2"]
        )
        with warnings.catch_warnings(record=True) as caught_warnings:
            result = _serialise_one_resource(res, AUTHOR_LOOKUP)
            assert len(caught_warnings) == 0
        expected = (
            b'<resource xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'label="lbl" id="id" restype=":Type">'
            b'<bitstream license="http://rdfh.ch/licenses/unknown" '
            b'copyright-holder="copy" '
            b'authorship-id="authorship_1"'
            b">"
            b"file.jpg"
            b"</bitstream>"
            b"</resource>"
        )
        assert etree.tostring(result) == expected

    def test_file_value_unknown_author(self) -> None:
        res = Resource.create_new("id", ":Type", "lbl").add_file(
            "file.jpg", LicenseRecommended.DSP.UNKNOWN, "copy", ["unknown"]
        )
        with warnings.catch_warnings(record=True) as caught_warnings:
            result = _serialise_one_resource(res, AUTHOR_LOOKUP)
            assert len(caught_warnings) == 1
        expected = (
            b'<resource xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'label="lbl" id="id" restype=":Type">'
            b'<bitstream license="http://rdfh.ch/licenses/unknown" copyright-holder="copy" authorship-id="unknown">'
            b"file.jpg"
            b"</bitstream>"
            b"</resource>"
        )
        assert etree.tostring(result) == expected

    def test_file_value_other_license(self) -> None:
        res = Resource.create_new("id", ":Type", "lbl").add_file(
            "file.jpg", LicenseOther.Various.BORIS_STANDARD, "copy", ["unknown"]
        )
        with warnings.catch_warnings(record=True) as caught_warnings:
            result = _serialise_one_resource(res, AUTHOR_LOOKUP)
            assert len(caught_warnings) == 1
        expected = (
            b'<resource xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'label="lbl" id="id" restype=":Type">'
            b'<bitstream license="http://rdfh.ch/licenses/boris" copyright-holder="copy" authorship-id="unknown">'
            b"file.jpg"
            b"</bitstream>"
            b"</resource>"
        )
        assert etree.tostring(result) == expected


class TestRegionResource:
    def test_serialise_no_warnings(self, region_no_warnings: RegionResource) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            _serialise_one_resource(region_no_warnings, AUTHOR_LOOKUP)
            assert len(caught_warnings) == 0

    def test_serialised_string_no_warnings(self, region_no_warnings: RegionResource) -> None:
        serialised = etree.tostring(_serialise_one_resource(region_no_warnings, AUTHOR_LOOKUP))
        expected = (
            b'<region xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'label="label" id="res_id">'
            b'<geometry-prop name="hasGeometry">'
            b'<geometry>{"status": "active", "type": "rectangle", "lineWidth": 2, '
            b'"points": [{"x": 0.1, "y": 0.1}, {"x": 0.2, "y": 0.2}]}</geometry>'
            b"</geometry-prop>"
            b'<color-prop name="hasColor"><color>#5b24bf</color></color-prop>'
            b'<resptr-prop name="isRegionOf"><resptr>region_of</resptr></resptr-prop>'
            b"</region>"
        )
        assert serialised == expected

    def test_serialise_no_region(self) -> None:
        region = RegionResource.create_new("res_id", "label", "region_of")
        with warnings.catch_warnings(record=True) as caught_warnings:
            _serialise_one_resource(region, AUTHOR_LOOKUP)
            assert len(caught_warnings) == 1


class TestLinkResource:
    def test_serialise_no_warnings(self, link_obj_no_warnings: LinkResource) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            _serialise_one_resource(link_obj_no_warnings, AUTHOR_LOOKUP)
            assert len(caught_warnings) == 0

    def test_serialised_string_no_warnings(self, link_obj_no_warnings: RegionResource) -> None:
        serialised = etree.tostring(_serialise_one_resource(link_obj_no_warnings, AUTHOR_LOOKUP))
        expected = (
            b'<link xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            b'label="lbl" id="id">'
            b'<resptr-prop name="hasLinkTo"><resptr>link</resptr></resptr-prop>'
            b'<text-prop name="hasComment"><text encoding="xml">cmt</text></text-prop>'
            b"</link>"
        )
        assert serialised == expected

    def test_serialise_no_comment(self) -> None:
        linkobj = LinkResource.create_new("id", "lbl", ["link"])
        with warnings.catch_warnings(record=True) as caught_warnings:
            _serialise_one_resource(linkobj, AUTHOR_LOOKUP)
            assert len(caught_warnings) == 1

    def test_serialise_no_link(self) -> None:
        with warnings.catch_warnings(record=True) as caught_warnings:
            linkobj = LinkResource.create_new("id", "lbl", []).add_comment("cmt")
            _serialise_one_resource(linkobj, AUTHOR_LOOKUP)
        warning_0 = regex.escape(
            "| Resource ID 'id' | Property 'hasLinkTo' | "
            "The input is empty. Please note that no values will be added to the resource."
        )
        warning_1 = regex.escape("A link object requires at least two links. Please note that an xmlupload will fail.")
        returned = sorted([x.message.args[0] for x in caught_warnings])  # type: ignore[union-attr]
        assert regex.search(warning_0, returned[0])
        assert regex.search(warning_1, returned[1])


if __name__ == "__main__":
    pytest.main([__file__])



from typing import Optional

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.project.legacy_models.group import Group
from dsp_tools.commands.project.legacy_models.project import Project
from dsp_tools.error.exceptions import BaseError


class ProjectContext:
    """Represents the project context"""

    _projects: list[Project]
    _inv_project_map: dict[str, str]  # dictionary of (project IRI:project name) pairs
    _groups: Optional[list[Group]]
    _group_map: Optional[dict[str, str]]
    _shortcode: Optional[str]
    _project_name: Optional[str]

    def __init__(self, con: Connection, shortcode: Optional[str] = None):
        self._shortcode = shortcode
        self._projects = Project.getAllProjects(con=con)
        self._inv_project_map: dict[str, str] = {x.iri: x.shortname for x in self._projects}
        try:
            self._groups = Group.getAllGroups(con=con)
        except BaseError:
            self._groups = None
        if self._groups:
            self._group_map: dict[str, str] = {
                f"{self._inv_project_map[x.project]}:{x.name}": x.iri for x in self._groups
            }
        else:
            self._group_map = None
        self._project_name = None
        # get the project name from the shortcode
        if self._shortcode:
            for p in self._projects:
                if p.shortcode == self._shortcode:
                    self._project_name = p.shortname
                    break

    @property
    def group_map(self) -> dict[str, str]:
        """Dictionary of (project:group name) and (group id) pairs of all groups in project"""
        return self._group_map

    @property
    def project_name(self) -> Optional[str]:
        """Name of the project"""
        return self._project_name



# mypy: disable-error-code="method-assign,no-untyped-def"

from copy import deepcopy

import pytest

from dsp_tools.commands.xmlupload.models.processed.res import ProcessedResource
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedSimpleText
from dsp_tools.commands.xmlupload.models.upload_state import UploadState
from dsp_tools.commands.xmlupload.upload_config import UploadConfig
from dsp_tools.commands.xmlupload.xmlupload import _tidy_up_resource_creation_idempotent


@pytest.fixture
def processed_resources() -> list[ProcessedResource]:
    one = ProcessedResource("foo_1_id", "onto:foo_1_type", "lbl", None, [ProcessedSimpleText("val", "prp", None, None)])
    two = ProcessedResource("foo_2_id", "onto:foo_2_type", "lbl", None, [ProcessedSimpleText("val", "prp", None, None)])
    return [one, two]


def test_idempotency_on_success(processed_resources) -> None:
    upload_state = UploadState(deepcopy(processed_resources), None, UploadConfig())
    for _ in range(3):
        _tidy_up_resource_creation_idempotent(upload_state, "foo_1_iri", processed_resources[0])
        assert upload_state.pending_resources == processed_resources[1:]
        assert upload_state.failed_uploads == []
        assert upload_state.iri_resolver.lookup == {"foo_1_id": "foo_1_iri"}
        assert not upload_state.pending_stash


def test_idempotency_on_failure(processed_resources) -> None:
    upload_state = UploadState(deepcopy(processed_resources), None, UploadConfig())
    for _ in range(3):
        _tidy_up_resource_creation_idempotent(upload_state, None, processed_resources[0])
        assert upload_state.pending_resources == processed_resources[1:]
        assert upload_state.failed_uploads == ["foo_1_id"]
        assert upload_state.iri_resolver.lookup == {}
        assert not upload_state.pending_stash



from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.permission import PermissionValue
from dsp_tools.error.xmllib_errors import XmllibFileNotFoundError
from dsp_tools.error.xmllib_errors import XmllibInputError
from dsp_tools.error.xmllib_errors import XmllibInternalError
from dsp_tools.utils.ansi_colors import BOLD_GREEN
from dsp_tools.utils.ansi_colors import YELLOW
from dsp_tools.utils.data_formats.date_util import is_full_date
from dsp_tools.utils.data_formats.shared import check_notna
from dsp_tools.utils.xml_parsing.parse_clean_validate_xml import validate_root_emit_user_message

PermissionValue.RV
PermissionValue.V
PermissionValue.M
PermissionValue.D
PermissionValue.CR

BOLD_GREEN
YELLOW

IriResolver.non_empty()

is_full_date("")

check_notna("")

validate_root_emit_user_message()

XmllibInputError()
XmllibInternalError()
XmllibFileNotFoundError()



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.utils.data_formats.date_util import Date


@dataclass
class IntervalFloats:
    start: float
    end: float


type ProcessedValueTypes = Union[bool, str, float, int, FormattedTextValue, Date, IntervalFloats]


@dataclass
class ProcessedValue:
    value: ProcessedValueTypes
    prop_iri: str
    comment: str | None
    permissions: Permissions | None


@dataclass
class ProcessedBoolean(ProcessedValue):
    value: bool


@dataclass
class ProcessedColor(ProcessedValue):
    value: str


@dataclass
class ProcessedDate(ProcessedValue):
    value: Date


@dataclass
class ProcessedDecimal(ProcessedValue):
    value: float


@dataclass
class ProcessedGeoname(ProcessedValue):
    value: str


@dataclass
class ProcessedGeometry(ProcessedValue):
    value: str


@dataclass
class ProcessedInt(ProcessedValue):
    value: int


@dataclass
class ProcessedInterval(ProcessedValue):
    value: IntervalFloats


@dataclass
class ProcessedLink(ProcessedValue):
    value: str
    value_uuid: str

    def __post_init__(self) -> None:
        self.prop_iri = f"{self.prop_iri}Value"


@dataclass
class ProcessedList(ProcessedValue):
    value: str


@dataclass
class ProcessedSimpleText(ProcessedValue):
    value: str


@dataclass
class ProcessedRichtext(ProcessedValue):
    value: FormattedTextValue
    resource_references: set[str]
    value_uuid: str


@dataclass
class ProcessedTime(ProcessedValue):
    value: str


@dataclass
class ProcessedUri(ProcessedValue):
    value: str



import shutil
import unittest
from pathlib import Path

import pytest

from dsp_tools.commands.template import generate_template_repo
from dsp_tools.error.exceptions import InputError

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)
# ruff: noqa: PT027 (pytest-unittest-raises-assertion) (remove this line when pytest is used instead of unittest)


class TestGenerateTemplates(unittest.TestCase):
    """Test the CLI command 'template'"""

    def test_generate_template_repo(self) -> None:
        """Test that the command succeeds the first time, fails the second time, and that the JSON + XML file exist."""
        success = generate_template_repo()
        self.assertTrue(success)

        with self.assertRaisesRegex(InputError, "already exists in your current working directory"):
            generate_template_repo()

        self.assertTrue(Path("0100-template-repo/template.json").exists())
        self.assertTrue(Path("0100-template-repo/template.xml").exists())

        shutil.rmtree("0100-template-repo")


if __name__ == "__main__":
    pytest.main([__file__])



from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from enum import StrEnum

from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.ansi_colors import YELLOW


class UserMessageSeverity(StrEnum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"


@dataclass
class MessageInfo:
    """
    message: message about what went wrong
    resource_id: ID of the affected resource
    prop_name: property name of the affected property (if applicable)
    field: information about which field of the resource is affected (if not the property), e.g. "resource id", "label"
    """

    message: str
    resource_id: str | None = None
    prop_name: str | None = None
    field: str | None = None


class XmllibUserInfoBase(Warning, ABC):
    """Base class for warnings that implement a custom showwarnings() function"""

    @classmethod
    @abstractmethod
    def showwarning(cls, message: str) -> None:
        """Functionality that should be executed when a warning of this class is emitted"""


class XmllibInputInfo(XmllibUserInfoBase):
    """If the xmllib input may be problematic"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        print(YELLOW + f"INFO | {message}" + RESET_TO_DEFAULT)


class XmllibInputWarning(XmllibUserInfoBase):
    """If the xmllib input is problematic"""

    @classmethod
    def showwarning(cls, message: str) -> None:
        print(BOLD_RED + f"WARNING | {message}" + RESET_TO_DEFAULT)



from __future__ import annotations

import json
from datetime import datetime
from pathlib import Path

from loguru import logger

from dsp_tools.commands.xmlupload.upload_config import DiagnosticsConfig


def write_id2iri_mapping(id2iri_mapping: dict[str, str], shortcode: str, diagnostics: DiagnosticsConfig) -> None:
    """Writes the mapping of internal IDs to IRIs to a file."""
    timestamp = datetime.now().strftime("%Y-%m-%d_%H%M%S")
    servername = diagnostics.server_as_foldername
    json_str = json.dumps(id2iri_mapping, ensure_ascii=False, indent=4)
    id2iri_filename_for_user = f"id2iri_{shortcode}_{servername}_{timestamp}.json"
    with open(id2iri_filename_for_user, "w", encoding="utf-8") as f:
        f.write(json_str)
        print(f"{datetime.now()}: The mapping of internal IDs to IRIs was written to {id2iri_filename_for_user}")
        logger.info(f"The mapping of internal IDs to IRIs was written to {id2iri_filename_for_user}")

    id_2_iri_folder = Path.home() / ".dsp-tools" / "id2iri"
    id_2_iri_folder.mkdir(parents=True, exist_ok=True)
    id2iri_filename_for_user_home = f"id2iri_{timestamp}_{shortcode}_{servername}.json"
    with open(id_2_iri_folder / id2iri_filename_for_user_home, "w", encoding="utf-8") as f:
        f.write(json_str)



from __future__ import annotations

import urllib
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Protocol

import requests
from loguru import logger
from requests import Session
from requests.adapters import HTTPAdapter
from requests.adapters import Retry

from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.commands.xmlupload.models.bitstream_info import BitstreamInfo
from dsp_tools.commands.xmlupload.models.processed.file_values import ProcessedFileValue
from dsp_tools.error.exceptions import BadCredentialsError
from dsp_tools.error.exceptions import InvalidIngestFileNameError
from dsp_tools.error.exceptions import PermanentConnectionError

STATUS_OK = 200
BAD_REQUEST = 400
STATUS_UNAUTHORIZED = 401
STATUS_INTERNAL_SERVER_ERROR = 500


@dataclass(frozen=True)
class IngestResponse:
    """Wrapper around `internal_filename`"""

    internal_filename: str


class AssetClient(Protocol):
    """Protocol for asset handling clients."""

    def get_bitstream_info(self, file_info: ProcessedFileValue) -> BitstreamInfo | None:
        """Uploads the file to the ingest server if applicable, and returns the upload results.

        Args:
            file_info: Information required for ingesting an asset
        """


@dataclass
class DspIngestClientLive(AssetClient):
    """Client for uploading assets to the DSP-Ingest."""

    dsp_ingest_url: str
    authentication_client: AuthenticationClient
    shortcode: str
    imgdir: str
    session: Session = field(init=False)

    def __post_init__(self) -> None:
        retries = 6
        self.session = Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=0.3,
            allowed_methods=None,  # means all methods
            status_forcelist=[STATUS_INTERNAL_SERVER_ERROR],
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)

    def _ingest(self, filepath: Path) -> IngestResponse:
        """Uploads a file to the ingest server and returns the IngestResponse.

        This function sends a POST request to the ingest server with the file to upload.
        It will retry the request 6 times in case of a connection, timeout or internal server error.

        After all retry attempts are exhausted it will raise exceptions if the upload failed.
        The http status code is also checked and if it is not 200, a PermanentConnectionError is raised.

        Args:
            filepath: Path to the file to ingest, could be either absolute or relative.

        Raises:
            BadCredentialsError: If the credentials are invalid.
            PermanentConnectionError: If the connection fails.

        Returns:
            IngestResponse: The internal filename of the uploaded file.
        """
        filename = urllib.parse.quote(filepath.name)
        url = f"{self.dsp_ingest_url}/projects/{self.shortcode}/assets/ingest/{filename}"
        headers = {
            "Authorization": f"Bearer {self.authentication_client.get_token()}",
            "Content-Type": "application/octet-stream",
        }
        timeout = 600
        with open(filepath, "rb") as binary_io:
            try:
                logger.debug(f"REQUEST: POST to {url}, timeout: {timeout}, headers: {headers | {'Authorization': '*'}}")
                res = self.session.post(
                    url=url,
                    headers=headers,
                    data=binary_io,
                    timeout=timeout,
                )
                logger.debug(f"RESPONSE: {res.status_code}: {res.text}")
                if res.status_code == STATUS_OK:
                    return IngestResponse(internal_filename=res.json()["internalFilename"])
                elif res.status_code == STATUS_UNAUTHORIZED:
                    raise BadCredentialsError("Bad credentials")
                elif res.status_code == BAD_REQUEST and res.text == "Invalid value for: path parameter filename":
                    raise InvalidIngestFileNameError()
                else:
                    raise PermanentConnectionError()
            except requests.exceptions.RequestException as e:
                raise PermanentConnectionError() from e

    def get_bitstream_info(self, file_info: ProcessedFileValue) -> BitstreamInfo | None:
        """Uploads a file to the ingest server and returns the upload results."""
        try:
            res = self._ingest(Path(self.imgdir) / Path(file_info.value))
            logger.info(f"Uploaded file '{file_info.value}'")
            return BitstreamInfo(res.internal_filename, file_info.metadata.permissions)
        except InvalidIngestFileNameError:
            msg = f"Invalid filename: Unable to upload file '{file_info.value}' of resource '{file_info.res_id}'"
        except PermanentConnectionError:
            msg = f"Unable to upload file '{file_info.value}' of resource '{file_info.res_id}'"
        logger.error(msg)
        return None


@dataclass(frozen=True)
class BulkIngestedAssetClient(AssetClient):
    """Client for handling media info, if the assets were bulk ingested previously."""

    def get_bitstream_info(self, file_info: ProcessedFileValue) -> BitstreamInfo:
        """Returns the BitstreamInfo of the already ingested file based on the `ProcessedFileValue.value`."""
        return BitstreamInfo(file_info.value, file_info.metadata.permissions)



from rdflib import XSD

from dsp_tools.commands.xmlupload.models.rdf_models import RDFPropTypeInfo
from dsp_tools.utils.rdflib_constants import KNORA_API

# paths for the validation files used by the docker container

ONTOLOGIES_DATA_TTL = "ONTOLOGIES_DATA.ttl"
ONTOLOGIES_SHACL_TTL = "ONTOLOGIES_SHACL.ttl"
ONTOLOGIES_REPORT_TTL = "ONTOLOGIES_REPORT.ttl"

CARDINALITY_DATA_TTL = "CARDINALITY_DATA.ttl"
CARDINALITY_SHACL_TTL = "CARDINALITY_SHACL.ttl"
CARDINALITY_REPORT_TTL = "CARDINALITY_REPORT.ttl"

CONTENT_DATA_TTL = "CONTENT_DATA.ttl"
CONTENT_SHACL_TTL = "CONTENT_SHACL.ttl"
CONTENT_REPORT_TTL = "CONTENT_REPORT.ttl"


ARCHIVE_FILE_VALUE = RDFPropTypeInfo(KNORA_API.ArchiveFileValue, KNORA_API.fileValueHasFilename, XSD.string)
AUDIO_FILE_VALUE = RDFPropTypeInfo(KNORA_API.AudioFileValue, KNORA_API.fileValueHasFilename, XSD.string)
DOCUMENT_FILE_VALUE = RDFPropTypeInfo(KNORA_API.DocumentFileValue, KNORA_API.fileValueHasFilename, XSD.string)
MOVING_IMAGE_FILE_VALUE = RDFPropTypeInfo(KNORA_API.MovingImageFileValue, KNORA_API.fileValueHasFilename, XSD.string)
STILL_IMAGE_FILE_VALUE = RDFPropTypeInfo(KNORA_API.StillImageFileValue, KNORA_API.fileValueHasFilename, XSD.string)
TEXT_FILE_VALUE = RDFPropTypeInfo(KNORA_API.TextFileValue, KNORA_API.fileValueHasFilename, XSD.string)
IIIF_URI_VALUE = RDFPropTypeInfo(
    KNORA_API.StillImageExternalFileValue, KNORA_API.stillImageFileValueHasExternalUrl, XSD.anyURI
)


# validation results

FILE_VALUE_PROPERTIES = {
    KNORA_API.hasArchiveFileValue,
    KNORA_API.hasAudioFileValue,
    KNORA_API.hasDocumentFileValue,
    KNORA_API.hasMovingImageFileValue,
    KNORA_API.hasTextFileValue,
    KNORA_API.hasStillImageFileValue,
    KNORA_API.hasLicense,
    KNORA_API.hasCopyrightHolder,
    KNORA_API.hasAuthorship,
}


FILEVALUE_DETAIL_INFO = {
    KNORA_API.hasArchiveFileValue: ("bitstream", "'zip', 'tar', 'gz', 'z', 'tgz', 'gzip', '7z'"),
    KNORA_API.hasAudioFileValue: ("bitstream", "'mp3', 'wav'"),
    KNORA_API.hasDocumentFileValue: ("bitstream", "'pdf', 'doc', 'docx', 'xls', 'xlsx', 'ppt', 'pptx', 'epub'"),
    KNORA_API.hasMovingImageFileValue: ("bitstream", "'mp4'"),
    KNORA_API.hasTextFileValue: ("bitstream", "'odd', 'rng', 'txt', 'xml', 'htm', 'html', 'xsd', 'xsl', 'csv', 'json'"),
    KNORA_API.hasStillImageFileValue: (
        "bitstream / iiif-uri",
        "'jpg', 'jpeg', 'png', 'tif', 'tiff', 'jp2' or a IIIF-URI",
    ),
}



from __future__ import annotations

from collections.abc import Collection
from dataclasses import dataclass
from pathlib import Path
from typing import Protocol

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_type_mismatch_warning
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning
from dsp_tools.utils.data_formats.uri_util import is_iiif_uri
from dsp_tools.xmllib.internal.checkers import check_and_warn_if_a_string_contains_a_potentially_empty_value
from dsp_tools.xmllib.internal.checkers import check_and_warn_potentially_empty_string
from dsp_tools.xmllib.internal.checkers import is_nonempty_value_internal
from dsp_tools.xmllib.internal.input_converters import check_and_fix_collection_input
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.licenses.recommended import License


@dataclass
class AuthorshipLookup:
    lookup: dict[tuple[str, ...], str]

    def get_id(self, authors: tuple[str, ...] | None) -> str | None:
        if authors is None:
            return None
        if not (found := self.lookup.get(authors)):
            emit_xmllib_input_warning(MessageInfo(f"The input authors {authors} are not defined in the look-up."))
            return " / ".join([str(x) for x in authors])
        return found


@dataclass
class Metadata:
    license: License | None
    copyright_holder: str | None
    authorship: tuple[str, ...] | None
    permissions: Permissions

    @classmethod
    def new(
        cls,
        license: License | None,
        copyright_holder: str | None,
        authorship: Collection[str] | None,
        permissions: Permissions,
        resource_id: str,
    ) -> Metadata:
        if license is not None and not isinstance(license, License):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="xmllib.License",
                value=license,
                res_id=resource_id,
                value_field="license (bistream/iiif-uri)",
            )
        if not isinstance(permissions, Permissions):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="xmllib.Permissions",
                value=permissions,
                res_id=resource_id,
                value_field="permissions (bistream/iiif-uri)",
            )
        if copyright_holder is not None:
            check_and_warn_potentially_empty_string(
                value=copyright_holder,
                res_id=resource_id,
                expected="string",
                field="copyright_holder (bistream/iiif-uri)",
            )
        if authorship is not None:
            if len(authorship) == 0:
                emit_xmllib_input_type_mismatch_warning(
                    expected_type="list of authorship strings",
                    value="empty input",
                    res_id=resource_id,
                    value_field="authorship (bistream/iiif-uri)",
                )
            fixed_authors = set(
                check_and_fix_collection_input(authorship, "authorship (bistream/iiif-uri)", resource_id)
            )
            for author in fixed_authors:
                check_and_warn_potentially_empty_string(
                    value=author, res_id=resource_id, expected="string", field="authorship (bistream/iiif-uri)"
                )
            fixed_authors_list = [str(x).strip() for x in fixed_authors]
            authors = tuple(sorted(fixed_authors_list))
        else:
            authors = None
        return cls(
            license=license,
            copyright_holder=copyright_holder,
            authorship=authors,
            permissions=permissions,
        )


class AbstractFileValue(Protocol):
    value: str
    metadata: Metadata
    comment: str | None


@dataclass
class FileValue(AbstractFileValue):
    value: str
    metadata: Metadata
    comment: str | None

    @classmethod
    def new(cls, value: str | Path, metadata: Metadata, comment: str | None, resource_id: str) -> FileValue:
        if isinstance(value, Path):
            if str(value) == ".":
                value = ""
            else:
                value = str(value)
        check_and_warn_potentially_empty_string(
            value=value,
            res_id=resource_id,
            expected="file name",
            field="bitstream",
        )
        if is_nonempty_value_internal(comment):
            fixed_comment = str(comment)
            check_and_warn_if_a_string_contains_a_potentially_empty_value(
                value=comment,
                res_id=resource_id,
                field="comment on bitstream",
            )
        else:
            fixed_comment = None
        return cls(value=str(value), metadata=metadata, comment=fixed_comment)


@dataclass
class IIIFUri(AbstractFileValue):
    value: str
    metadata: Metadata
    comment: str | None

    @classmethod
    def new(cls, value: str, metadata: Metadata, comment: str | None, resource_id: str) -> IIIFUri:
        if not is_iiif_uri(value):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="IIIF uri",
                value=value,
                res_id=resource_id,
            )
        if is_nonempty_value_internal(comment):
            fixed_comment = str(comment)
            check_and_warn_if_a_string_contains_a_potentially_empty_value(
                value=comment,
                res_id=resource_id,
                field="comment on iiif-uri",
            )
        else:
            fixed_comment = None
        return cls(value=str(value), metadata=metadata, comment=fixed_comment)



"""end to end tests for group class"""

import unittest
from typing import cast

import pytest

from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.clients.connection import Connection
from dsp_tools.clients.connection_live import ConnectionLive
from dsp_tools.commands.project.legacy_models.group import Group
from dsp_tools.legacy_models.langstring import LangString
from dsp_tools.legacy_models.langstring import Languages

# ruff: noqa: PT009 (pytest-unittest-assertion) (remove this line when pytest is used instead of unittest)


class TestGroup(unittest.TestCase):
    test_project = "http://rdfh.ch/projects/0001"
    con: Connection

    def setUp(self) -> None:
        """
        Creates a connection to DSP-API.
        For each test method, a new TestCase instance is created, so setUp() is executed before each test method.
        """
        auth = AuthenticationClientLive("http://0.0.0.0:3333", "root@example.com", "test")
        self.con = ConnectionLive("http://0.0.0.0:3333", auth)

    def test_group_getAllGroups(self) -> None:
        """
        Retrieve all groups
        :return: None
        """
        groups = Group.getAllGroups(self.con)
        group_ids = [group.iri for group in groups]
        group_ids_expected = [
            "http://rdfh.ch/groups/00FF/images-reviewer",
            "http://rdfh.ch/groups/0001/thing-searcher",
        ]
        self.assertTrue(set(group_ids).issuperset(set(group_ids_expected)))

    def test_group_create(self) -> None:
        """
        Create a group
        :return: None
        """
        group = Group(
            con=self.con,
            name="Group create",
            descriptions=LangString({Languages.EN: "This is group create"}),
            project=self.test_project,
            status=True,
            selfjoin=False,
        )

        self.assertEqual(group.name, "Group create")
        self.assertCountEqual(
            cast(list[dict[str, str]], group.descriptions.toJsonObj()),
            [{"language": "en", "value": "This is group create"}],
        )
        self.assertEqual(group.project, self.test_project)
        self.assertTrue(group.status)
        self.assertFalse(group.selfjoin)

    def test_group_read(self) -> None:
        """
        Read an existing group
        :return: None
        """
        group = Group(con=self.con, iri="http://rdfh.ch/groups/0001/thing-searcher").read()
        self.assertEqual(group.name, "Thing searcher")
        self.assertEqual(group.project, self.test_project)
        self.assertTrue(group.status)
        self.assertTrue(group.selfjoin)


if __name__ == "__main__":
    pytest.main([__file__])



from typing import cast

from loguru import logger
from rdflib import RDF
from rdflib import RDFS
from rdflib import SH
from rdflib import XSD
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.validate_data.constants import FILE_VALUE_PROPERTIES
from dsp_tools.commands.validate_data.constants import FILEVALUE_DETAIL_INFO
from dsp_tools.commands.validate_data.mappers import RESULT_TO_PROBLEM_MAPPER
from dsp_tools.commands.validate_data.models.input_problems import AllProblems
from dsp_tools.commands.validate_data.models.input_problems import InputProblem
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import Severity
from dsp_tools.commands.validate_data.models.validation import DetailBaseInfo
from dsp_tools.commands.validate_data.models.validation import QueryInfo
from dsp_tools.commands.validate_data.models.validation import ReformattedIRI
from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.models.validation import ValidationResult
from dsp_tools.commands.validate_data.models.validation import ValidationResultBaseInfo
from dsp_tools.commands.validate_data.models.validation import ViolationType
from dsp_tools.commands.validate_data.utils import reformat_any_iri
from dsp_tools.commands.validate_data.utils import reformat_data_iri
from dsp_tools.commands.validate_data.utils import reformat_onto_iri
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.rdflib_constants import API_SHAPES
from dsp_tools.utils.rdflib_constants import DASH
from dsp_tools.utils.rdflib_constants import KNORA_API
from dsp_tools.utils.rdflib_constants import SubjectObjectTypeAlias

LEGAL_INFO_PROPS = {KNORA_API.hasLicense, KNORA_API.hasCopyrightHolder, KNORA_API.hasAuthorship}


SEVERITY_MAPPER: dict[SubjectObjectTypeAlias, Severity] = {
    SH.Violation: Severity.VIOLATION,
    SH.Warning: Severity.WARNING,
    SH.Info: Severity.INFO,
}


def reformat_validation_graph(report: ValidationReportGraphs) -> AllProblems:
    """
    Reformats the validation result from an RDF graph into class instances
    that are used to communicate the problems with the user.

    Args:
        report: with all the information necessary to construct a user message

    Returns:
        All Problems
    """
    logger.debug("Reformatting validation results.")
    results_and_onto = report.validation_graph + report.onto_graph
    data_and_onto = report.onto_graph + report.data_graph
    validation_results, unexpected_extracted = _query_all_results(results_and_onto, data_and_onto)
    reformatted_results = _reformat_extracted_results(validation_results)
    return AllProblems(reformatted_results, unexpected_extracted)


def _query_all_results(
    results_and_onto: Graph, data_onto_graph: Graph
) -> tuple[list[ValidationResult], list[UnexpectedComponent]]:
    no_details, with_details = _separate_result_types(results_and_onto, data_onto_graph)
    extracted_results: list[ValidationResult] = []
    unexpected_components: list[UnexpectedComponent] = []

    no_detail_extracted, no_detail_unexpected = _query_all_without_detail(no_details, results_and_onto, data_onto_graph)
    extracted_results.extend(no_detail_extracted)
    unexpected_components.extend(no_detail_unexpected)

    detail_reformatted, detail_unexpected = _query_all_with_detail(with_details, results_and_onto, data_onto_graph)
    extracted_results.extend(detail_reformatted)
    unexpected_components.extend(detail_unexpected)
    return extracted_results, unexpected_components


def _separate_result_types(
    results_and_onto: Graph, data_onto_graph: Graph
) -> tuple[list[ValidationResultBaseInfo], list[ValidationResultBaseInfo]]:
    base_info = _extract_base_info_of_resource_results(results_and_onto, data_onto_graph)
    no_details = [x for x in base_info if not x.detail]
    with_details = [x for x in base_info if x.detail]
    return no_details, with_details


def _extract_base_info_of_resource_results(
    results_and_onto: Graph, data_onto_graph: Graph
) -> list[ValidationResultBaseInfo]:
    all_res_focus_nodes = []
    main_bns = _get_all_main_result_bns(results_and_onto)
    value_types = _get_all_value_classes(data_onto_graph)
    for nd in main_bns:
        focus_iri = next(results_and_onto.objects(nd, SH.focusNode))
        res_type = next(data_onto_graph.objects(focus_iri, RDF.type))
        info = QueryInfo(
            validation_bn=nd,
            focus_iri=focus_iri,
            focus_rdf_type=res_type,
        )
        all_res_focus_nodes.extend(_extract_one_base_info(info, results_and_onto, data_onto_graph, value_types))
    return all_res_focus_nodes


def _get_all_value_classes(data_onto_graph: Graph) -> set[SubjectObjectTypeAlias]:
    all_types = set(data_onto_graph.objects(predicate=RDF.type))
    all_value_types = set()
    for type_ in all_types:
        super_classes = set(data_onto_graph.transitive_objects(type_, RDFS.subClassOf))
        if KNORA_API.Value in super_classes:
            all_value_types.add(type_)
    return all_value_types


def _extract_one_base_info(
    info: QueryInfo, results_and_onto: Graph, data_onto_graph: Graph, value_types: set[SubjectObjectTypeAlias]
) -> list[ValidationResultBaseInfo]:
    results = []
    path = next(results_and_onto.objects(info.validation_bn, SH.resultPath))
    main_component_type = next(results_and_onto.objects(info.validation_bn, SH.sourceConstraintComponent))
    severity = next(results_and_onto.objects(info.validation_bn, SH.resultSeverity))
    if detail_bn_list := list(results_and_onto.objects(info.validation_bn, SH.detail)):
        for single_detail in detail_bn_list:
            detail_component = next(results_and_onto.objects(single_detail, SH.sourceConstraintComponent))
            detail = DetailBaseInfo(
                detail_bn=single_detail,
                source_constraint_component=detail_component,
            )
            results.append(
                ValidationResultBaseInfo(
                    result_bn=info.validation_bn,
                    source_constraint_component=main_component_type,
                    focus_node_iri=info.focus_iri,
                    focus_node_type=info.focus_rdf_type,
                    result_path=path,
                    severity=severity,
                    detail=detail,
                )
            )
    else:
        resource_iri, resource_type, user_facing_prop = _get_resource_iri_and_type(
            info, path, data_onto_graph, value_types
        )
        results.append(
            ValidationResultBaseInfo(
                result_bn=info.validation_bn,
                source_constraint_component=main_component_type,
                focus_node_iri=resource_iri,
                focus_node_type=resource_type,
                result_path=user_facing_prop,
                severity=severity,
                detail=None,
            )
        )
    return results


def _get_all_main_result_bns(results_and_onto: Graph) -> set[SubjectObjectTypeAlias]:
    all_bns = set(results_and_onto.subjects(RDF.type, SH.ValidationResult))
    # All the blank nodes that are referenced in a sh:detail will be queried together with the main validation result
    # if we queried them separately we would get duplicate errors
    detail_bns = set(results_and_onto.objects(predicate=SH.detail))
    return all_bns - detail_bns


def _get_resource_iri_and_type(
    info: QueryInfo, path: SubjectObjectTypeAlias, data_onto_graph: Graph, value_types: set[SubjectObjectTypeAlias]
) -> tuple[SubjectObjectTypeAlias, SubjectObjectTypeAlias, SubjectObjectTypeAlias]:
    resource_iri, resource_type, user_facing_prop = info.focus_iri, info.focus_rdf_type, path
    if info.focus_rdf_type in value_types:
        resource_iri, predicate = next(data_onto_graph.subject_predicates(object=info.focus_iri))
        resource_type = next(data_onto_graph.objects(resource_iri, RDF.type))
        if user_facing_prop not in LEGAL_INFO_PROPS:
            user_facing_prop = predicate
    return resource_iri, resource_type, user_facing_prop


def _query_all_without_detail(
    all_base_info: list[ValidationResultBaseInfo], results_and_onto: Graph, data: Graph
) -> tuple[list[ValidationResult], list[UnexpectedComponent]]:
    extracted_results: list[ValidationResult] = []
    unexpected_components: list[UnexpectedComponent] = []

    for base_info in all_base_info:
        res = _query_one_without_detail(base_info, results_and_onto, data)
        if res is None:
            pass
        elif isinstance(res, UnexpectedComponent):
            unexpected_components.append(res)
        else:
            extracted_results.append(res)
    return extracted_results, unexpected_components


def _query_one_without_detail(  # noqa:PLR0911 (Too many return statements)
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data: Graph
) -> ValidationResult | UnexpectedComponent | None:
    msg = next(results_and_onto.objects(base_info.result_bn, SH.resultMessage))
    component = next(results_and_onto.objects(base_info.result_bn, SH.sourceConstraintComponent))
    match component:
        case SH.PatternConstraintComponent:
            return _query_pattern_constraint_component_violation(base_info.result_bn, base_info, results_and_onto)
        case SH.MinCountConstraintComponent:
            return _query_for_min_cardinality_violation(base_info, msg)
        case SH.MaxCountConstraintComponent:
            return ValidationResult(
                violation_type=ViolationType.MAX_CARD,
                res_iri=base_info.focus_node_iri,
                res_class=base_info.focus_node_type,
                severity=base_info.severity,
                property=base_info.result_path,
                expected=msg,
            )
        case SH.LessThanOrEqualsConstraintComponent:
            return _query_for_less_than_or_equal_violation(base_info, results_and_onto, data, msg)
        case DASH.ClosedByTypesConstraintComponent:
            return _query_for_non_existent_cardinality_violation(base_info, results_and_onto, data)
        case SH.SPARQLConstraintComponent:
            return _query_for_unique_value_violation(base_info, results_and_onto)
        case DASH.CoExistsWithConstraintComponent:
            return _query_for_coexists_with_violation(base_info, results_and_onto, data, msg)
        case SH.ClassConstraintComponent:
            return _query_class_constraint_without_detail(base_info, results_and_onto, data, msg)
        case (
            SH.InConstraintComponent
            | SH.LessThanConstraintComponent
            | SH.MinExclusiveConstraintComponent
            | SH.MinInclusiveConstraintComponent
            | DASH.SingleLineConstraintComponent
        ):
            return _query_general_violation_info(
                base_info.result_bn, base_info, results_and_onto, ViolationType.GENERIC
            )
        case SH.OrConstraintComponent:
            return _query_general_violation_info_with_value_as_string(
                base_info.result_bn, base_info, results_and_onto, data
            )
        case _:
            return UnexpectedComponent(str(component))


def _query_class_constraint_without_detail(
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data: Graph, message: SubjectObjectTypeAlias
) -> ValidationResult | None:
    val: None | SubjectObjectTypeAlias = next(results_and_onto.objects(base_info.result_bn, SH.value))
    # In this case we have some kind of FileValue violation
    violation_type = ViolationType.GENERIC
    value_type: None | SubjectObjectTypeAlias = None
    msg: None | SubjectObjectTypeAlias = message
    expected = None
    val_type_list = list(data.objects(val, RDF.type))
    # Here we have a normal value type violation
    if val_type_list:
        val_type = val_type_list.pop(0)
        value_super_class = list(results_and_onto.transitive_objects(val_type, RDFS.subClassOf))
        if KNORA_API.FileValue not in value_super_class:
            value_type = val_type
            val = None
            violation_type = ViolationType.VALUE_TYPE
            msg = None
            expected = message
    return ValidationResult(
        violation_type=violation_type,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        message=msg,
        expected=expected,
        input_value=val,
        input_type=value_type,
    )


def _query_for_less_than_or_equal_violation(
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data: Graph, message: SubjectObjectTypeAlias
) -> ValidationResult | None:
    value_iri = next(results_and_onto.objects(base_info.result_bn, SH.focusNode))
    start = next(data.objects(value_iri, API_SHAPES.dateHasStart))
    end = next(data.objects(value_iri, API_SHAPES.dateHasEnd))
    start_lit = cast(Literal, start)
    end_lit = cast(Literal, end)
    start_is_string = start_lit.datatype == XSD.string
    end_is_string = end_lit.datatype == XSD.string
    # If any one of the date ranges cannot be parsed as an xsd date, we get this violation also.
    # But the main problem is, that the date format is wrong, in which case the datatype is xsd:string.
    # This produces its own message
    if any([start_is_string, end_is_string]):
        return None
    return ValidationResult(
        violation_type=ViolationType.GENERIC,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        input_value=_get_value_as_string(base_info.result_bn, results_and_onto, data),
        message=message,
    )


def _query_for_non_existent_cardinality_violation(
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data: Graph
) -> ValidationResult | None:
    input_val = None
    if base_info.result_path in FILE_VALUE_PROPERTIES:
        violation_type = ViolationType.FILE_VALUE_PROHIBITED
        if value_bn_found := list(results_and_onto.objects(base_info.result_bn, SH.value)):
            value_bn = value_bn_found.pop(0)
            if file_path := list(data.objects(value_bn, KNORA_API.fileValueHasFilename)):
                input_val = file_path.pop(0)
            elif iiif_uri := list(data.objects(value_bn, KNORA_API.stillImageFileValueHasExternalUrl)):
                input_val = iiif_uri.pop(0)
    else:
        violation_type = ViolationType.NON_EXISTING_CARD
    return ValidationResult(
        violation_type=violation_type,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        input_value=input_val,
    )


def _query_all_with_detail(
    all_base_info: list[ValidationResultBaseInfo], results_and_onto: Graph, data_onto_graph: Graph
) -> tuple[list[ValidationResult], list[UnexpectedComponent]]:
    extracted_results: list[ValidationResult] = []
    unexpected_components: list[UnexpectedComponent] = []

    for base_info in all_base_info:
        res = _query_one_with_detail(base_info, results_and_onto, data_onto_graph)
        if isinstance(res, UnexpectedComponent):
            unexpected_components.append(res)
        else:
            extracted_results.append(res)
    return extracted_results, unexpected_components


def _query_one_with_detail(
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data_graph: Graph
) -> ValidationResult | UnexpectedComponent:
    detail_info = cast(DetailBaseInfo, base_info.detail)
    match detail_info.source_constraint_component:
        case SH.MinCountConstraintComponent:
            if base_info.result_path in FILE_VALUE_PROPERTIES:
                return _query_general_violation_info(
                    base_info.result_bn, base_info, results_and_onto, ViolationType.GENERIC
                )
            return _query_for_value_type_violation(base_info, results_and_onto, data_graph)
        case SH.PatternConstraintComponent:
            return _query_pattern_constraint_component_violation(detail_info.detail_bn, base_info, results_and_onto)
        case SH.ClassConstraintComponent:
            return _query_class_constraint_component_violation(base_info, results_and_onto, data_graph)
        case SH.InConstraintComponent | DASH.SingleLineConstraintComponent:
            detail = cast(DetailBaseInfo, base_info.detail)
            return _query_general_violation_info(detail.detail_bn, base_info, results_and_onto, ViolationType.GENERIC)
        case _:
            return UnexpectedComponent(str(detail_info.source_constraint_component))


def _query_class_constraint_component_violation(
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data_graph: Graph
) -> ValidationResult | UnexpectedComponent:
    detail_info = cast(DetailBaseInfo, base_info.detail)
    detail_path = next(results_and_onto.objects(detail_info.detail_bn, SH.resultPath))
    if detail_path == RDF.type:
        return _query_for_value_type_violation(base_info, results_and_onto, data_graph)
    return _query_for_link_value_target_violation(base_info, results_and_onto, data_graph)


def _query_for_value_type_violation(
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data_graph: Graph
) -> ValidationResult:
    detail_info = cast(DetailBaseInfo, base_info.detail)
    msg = next(results_and_onto.objects(detail_info.detail_bn, SH.resultMessage))
    val = next(results_and_onto.objects(base_info.result_bn, SH.value))
    val_type = next(data_graph.objects(val, RDF.type))
    return ValidationResult(
        violation_type=ViolationType.VALUE_TYPE,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        expected=msg,
        input_type=val_type,
    )


def _query_pattern_constraint_component_violation(
    bn_with_info: SubjectObjectTypeAlias, base_info: ValidationResultBaseInfo, results_and_onto: Graph
) -> ValidationResult:
    val = next(results_and_onto.objects(bn_with_info, SH.value))
    msg = next(results_and_onto.objects(bn_with_info, SH.resultMessage))
    return ValidationResult(
        violation_type=ViolationType.PATTERN,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        expected=msg,
        input_value=val,
    )


def _query_general_violation_info(
    result_bn: SubjectObjectTypeAlias,
    base_info: ValidationResultBaseInfo,
    results_and_onto: Graph,
    violation_type: ViolationType,
    value: SubjectObjectTypeAlias | None = None,
) -> ValidationResult:
    if not value:
        val = next(results_and_onto.objects(result_bn, SH.value), None)
    else:
        val = value
    msg = next(results_and_onto.objects(result_bn, SH.resultMessage))
    return ValidationResult(
        violation_type=violation_type,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        message=msg,
        input_value=val,
    )


def _query_general_violation_info_with_value_as_string(
    result_bn: SubjectObjectTypeAlias,
    base_info: ValidationResultBaseInfo,
    results_and_onto: Graph,
    data: Graph,
) -> ValidationResult:
    value_iri = next(results_and_onto.objects(result_bn, SH.focusNode))
    value = next(data.objects(value_iri, KNORA_API.valueAsString))
    return _query_general_violation_info(result_bn, base_info, results_and_onto, ViolationType.GENERIC, value)


def _query_for_link_value_target_violation(
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data_graph: Graph
) -> ValidationResult:
    detail_info = cast(DetailBaseInfo, base_info.detail)
    target_iri = next(results_and_onto.objects(detail_info.detail_bn, SH.value))
    target_rdf_type = next(data_graph.objects(target_iri, RDF.type), None)
    expected_type = next(results_and_onto.objects(detail_info.detail_bn, SH.resultMessage))
    return ValidationResult(
        violation_type=ViolationType.LINK_TARGET,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        expected=expected_type,
        input_value=target_iri,
        input_type=target_rdf_type,
    )


def _query_for_min_cardinality_violation(
    base_info: ValidationResultBaseInfo, msg: SubjectObjectTypeAlias
) -> ValidationResult:
    if base_info.result_path in LEGAL_INFO_PROPS:
        violation_type = ViolationType.GENERIC
    else:
        violation_type = ViolationType.MIN_CARD
    return ValidationResult(
        violation_type=violation_type,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        expected=msg,
    )


def _query_for_unique_value_violation(
    base_info: ValidationResultBaseInfo,
    results_and_onto: Graph,
) -> ValidationResult:
    val = next(results_and_onto.objects(base_info.result_bn, SH.value))
    return ValidationResult(
        violation_type=ViolationType.UNIQUE_VALUE,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        severity=base_info.severity,
        property=base_info.result_path,
        input_value=val,
    )


def _query_for_coexists_with_violation(
    base_info: ValidationResultBaseInfo, results_and_onto: Graph, data: Graph, message: SubjectObjectTypeAlias
) -> ValidationResult:
    source_shapes = next(results_and_onto.objects(base_info.result_bn, SH.sourceShape))
    if source_shapes == API_SHAPES.seqnum_PropShape:
        violation_type = ViolationType.SEQNUM_IS_PART_OF
        value = None
        prop = None
    else:
        violation_type = ViolationType.GENERIC
        value = _get_value_as_string(base_info.result_bn, results_and_onto, data)
        prop = base_info.result_path
    return ValidationResult(
        violation_type=violation_type,
        res_iri=base_info.focus_node_iri,
        res_class=base_info.focus_node_type,
        property=prop,
        severity=base_info.severity,
        message=message,
        input_value=value,
    )


def _get_value_as_string(result_bn: SubjectObjectTypeAlias, results: Graph, data: Graph) -> SubjectObjectTypeAlias:
    value_iri = next(results.objects(result_bn, SH.focusNode))
    return next(data.objects(value_iri, KNORA_API.valueAsString))


def _reformat_extracted_results(results: list[ValidationResult]) -> list[InputProblem]:
    return [_reformat_one_validation_result(x) for x in results]


def _reformat_one_validation_result(validation_result: ValidationResult) -> InputProblem:
    match validation_result.violation_type:
        case ViolationType.MIN_CARD:
            return _reformat_min_card(validation_result)
        case (
            ViolationType.MAX_CARD
            | ViolationType.NON_EXISTING_CARD
            | ViolationType.PATTERN
            | ViolationType.UNIQUE_VALUE
            | ViolationType.VALUE_TYPE as violation
        ):
            problem = RESULT_TO_PROBLEM_MAPPER[violation]
            return _reformat_generic(result=validation_result, problem_type=problem)
        case ViolationType.GENERIC:
            prop_str = None
            if validation_result.property in LEGAL_INFO_PROPS or validation_result.property in FILE_VALUE_PROPERTIES:
                prop_str = "bitstream / iiif-uri"
            return _reformat_generic(validation_result, ProblemType.GENERIC, prop_string=prop_str)
        case ViolationType.FILE_VALUE_PROHIBITED | ViolationType.FILE_VALUE_MISSING as violation:
            problem = RESULT_TO_PROBLEM_MAPPER[violation]
            return _reformat_generic(result=validation_result, problem_type=problem, prop_string="bitstream / iiif-uri")
        case ViolationType.SEQNUM_IS_PART_OF:
            return _reformat_generic(
                result=validation_result, problem_type=ProblemType.GENERIC, prop_string="seqnum or isPartOf"
            )
        case ViolationType.LINK_TARGET:
            return _reformat_link_target_violation_result(validation_result)
        case _:
            raise BaseError(f"An unknown violation result was found: {validation_result.__class__.__name__}")


def _reformat_min_card(result: ValidationResult) -> InputProblem:
    iris = _reformat_main_iris(result)
    if file_prop_info := FILEVALUE_DETAIL_INFO.get(cast(URIRef, result.property)):
        prop_str, file_extensions = file_prop_info
        detail_msg = None
        problem_type = ProblemType.FILE_VALUE_MISSING
        expected: str | None = f"This resource requires a file with one of the following extensions: {file_extensions}"
    else:
        prop_str = iris.prop_name
        detail_msg = _convert_rdflib_input_to_string(result.message)
        problem_type = ProblemType.MIN_CARD
        expected = _convert_rdflib_input_to_string(result.expected)

    return InputProblem(
        problem_type=problem_type,
        res_id=iris.res_id,
        res_type=iris.res_type,
        prop_name=prop_str,
        severity=SEVERITY_MAPPER[result.severity],
        message=detail_msg,
        input_value=_convert_rdflib_input_to_string(result.input_value),
        input_type=_convert_rdflib_input_to_string(result.input_type),
        expected=expected,
    )


def _reformat_generic(
    result: ValidationResult, problem_type: ProblemType, prop_string: str | None = None
) -> InputProblem:
    iris = _reformat_main_iris(result)
    user_prop = iris.prop_name if not prop_string else prop_string
    return InputProblem(
        problem_type=problem_type,
        res_id=iris.res_id,
        res_type=iris.res_type,
        prop_name=user_prop,
        severity=SEVERITY_MAPPER[result.severity],
        message=_convert_rdflib_input_to_string(result.message),
        input_value=_convert_rdflib_input_to_string(result.input_value),
        input_type=_convert_rdflib_input_to_string(result.input_type),
        expected=_convert_rdflib_input_to_string(result.expected),
    )


def _reformat_link_target_violation_result(result: ValidationResult) -> InputProblem:
    iris = _reformat_main_iris(result)
    input_type = None
    expected = None
    problem_type = ProblemType.INEXISTENT_LINKED_RESOURCE

    if result.input_type:
        problem_type = ProblemType.LINK_TARGET_TYPE_MISMATCH
        input_type = reformat_onto_iri(str(result.input_type))
        expected = reformat_onto_iri(str(result.expected))

    return InputProblem(
        problem_type=problem_type,
        res_id=iris.res_id,
        res_type=iris.res_type,
        prop_name=iris.prop_name,
        severity=SEVERITY_MAPPER[result.severity],
        input_value=reformat_data_iri(str(result.input_value)),
        input_type=input_type,
        expected=expected,
    )


def _reformat_main_iris(result: ValidationResult) -> ReformattedIRI:
    subject_id = reformat_data_iri(result.res_iri)
    prop_name = reformat_onto_iri(result.property) if result.property else ""
    res_type = reformat_onto_iri(result.res_class)
    return ReformattedIRI(res_id=subject_id, res_type=res_type, prop_name=prop_name)


def _convert_rdflib_input_to_string(input_val: SubjectObjectTypeAlias | None) -> str | None:
    if not input_val:
        return None
    if isinstance(input_val, URIRef):
        return reformat_any_iri(input_val)
    return str(input_val)



# mypy: disable-error-code="no-untyped-def"

import warnings
from typing import Any

import numpy as np
import pandas as pd
import polars as pl
import pytest

from dsp_tools.xmllib.internal.constants import KNOWN_XML_TAGS
from dsp_tools.xmllib.value_checkers import check_richtext_syntax
from dsp_tools.xmllib.value_checkers import is_bool_like
from dsp_tools.xmllib.value_checkers import is_color
from dsp_tools.xmllib.value_checkers import is_date
from dsp_tools.xmllib.value_checkers import is_decimal
from dsp_tools.xmllib.value_checkers import is_dsp_ark
from dsp_tools.xmllib.value_checkers import is_dsp_iri
from dsp_tools.xmllib.value_checkers import is_geoname
from dsp_tools.xmllib.value_checkers import is_integer
from dsp_tools.xmllib.value_checkers import is_link_value
from dsp_tools.xmllib.value_checkers import is_nonempty_value
from dsp_tools.xmllib.value_checkers import is_timestamp
from dsp_tools.xmllib.value_checkers import is_valid_resource_id


@pytest.mark.parametrize(
    "text",
    [
        "None",
        "a",
        "-",
        0,
        1,
        "0",
        "1",
        True,
        False,
        "עִבְרִית",
        "اَلْعَرَبِيَّةُ",
    ],
)
def test_is_nonempty_value_correct(text: str) -> None:
    assert is_nonempty_value(text)


@pytest.mark.parametrize(
    "text",
    [
        np.nan,
        pd.NA,
        None,
        "",
        " ",
        " \t ",
        " \n ",
        " \r ",
        " \v ",  # vertical tab
        " \f ",
        " \u00a0 ",  # non-breaking space
        " \u200b ",  # zero-width space
        " \ufeff ",  # Zero-Width No-Break Space
        " \t\n\r\v\f \u00a0 \u200b \ufeff",
    ],
)
def test_is_nonempty_value_wrong(text: str) -> None:
    assert not is_nonempty_value(text)


def test_is_nonempty_value_wrong_polars_types() -> None:
    df = pl.DataFrame({"col": [float("nan"), np.nan, None]})
    for row in df.iter_rows(named=True):
        assert not is_nonempty_value(row["col"])


@pytest.mark.parametrize(
    "val", ["false", "0", "0.0", "no", "true", "1", "1.0", "yes", False, True, "oui", "JA", "non", "nein"]
)
def test_is_bool_correct(val: Any) -> None:
    assert is_bool_like(val)


@pytest.mark.parametrize("val", ["none", "asdfioh", None, pd.NA])
def test_is_bool_wrong(val: Any) -> None:
    assert not is_bool_like(val)


@pytest.mark.parametrize("val", ["#234567", "#abcdef", "#123abc"])
def test_is_color_correct(val: str) -> None:
    assert is_color(val)


@pytest.mark.parametrize("val", [12334, "123abc", "234567#"])
def test_is_color_wrong(val: Any) -> None:
    assert not is_color(val)


@pytest.mark.parametrize(
    "val",
    [
        "GREGORIAN:CE:0476-09-04:CE:0476-09-04",
        "GREGORIAN:BC:2000:BC:1000",
        "JULIAN:CE:0476-09-04:CE:0476-09-04",
        "JULIAN:BC:2000:BC:1000",
        "ISLAMIC:0476-09-04:0476-09-04",
        "ISLAMIC:1000:2000",
        "GREGORIAN:0476-09-04:0476-09-04",
        "CE:0476-09-04:CE:0476-09-04",
    ],
)
def test_is_date_correct(val: str) -> None:
    assert is_date(val)


@pytest.mark.parametrize(
    "val",
    [
        "OTHER:BC:2000:BC:1000",
        "GREGORIAN:other:2000:other:1000",
        "GREGORIAN::CE:0476-09-04:CE:0476-09-04",
        "GREGORIAN:CE:0476-09-04:CE::0476-09-04",
        "GREGORIAN:CE:0476-09-04::CE:0476-09-04",
        "GREGORIAN:CE::0476-09-04:CE:0476-09-04",
        "ISLAMIC:BC:0476-09-04:BC:0476-09-04",
        "ISLAMIC:BCE:0476-09-04:BCE:0476-09-04",
        "ISLAMIC:CE:0476-09-04:CE:0476-09-04",
        "ISLAMIC:AD:0476-09-04:AD:0476-09-04",
        "ISLAMIC:AH:0476-09-04:AH:0476-09-04",  # Anno Hegirae is not implemented yet
        "ISLAMIC:BH:0476-09-04:BH:0476-09-04",  # Before Hijra is not implemented yet
        "ISLAMIC:AH:0476-09-04:0476-09-04",
        "ISLAMIC:0476-09-04:AH:0476-09-04",
    ],
)
def test_is_date_wrong(val: Any) -> None:
    assert not is_date(val)


@pytest.mark.parametrize("val", ["00001", 23412])
def test_is_geoname_correct(val: Any) -> None:
    assert is_geoname(val)


@pytest.mark.parametrize("val", [122.2, "asdf"])
def test_is_geoname_wrong(val: Any) -> None:
    assert not is_geoname(val)


@pytest.mark.parametrize("val", [1.2, "1.432", 1, "1", -1.1, "1e-1", "1e2"])
def test_is_decimal_correct(val: Any) -> None:
    assert is_decimal(val)


@pytest.mark.parametrize("val", ["asf", np.nan, pd.NA, True, False])
def test_is_decimal_wrong(val: Any) -> None:
    assert not is_decimal(val)


@pytest.mark.parametrize("val", [1, "1", -1])
def test_is_integer_correct(val: Any) -> None:
    assert is_integer(val)


@pytest.mark.parametrize("val", [1.2, "1.2", "wdasd", True, False, "1e2"])
def test_is_integer_wrong(val: Any) -> None:
    assert not is_integer(val)


@pytest.mark.parametrize("val", ["_1", "abc_2", "http://rdfh.ch/4123/54SYvWF0QUW6a"])
def test_is_link_value_correct(val):
    assert is_link_value(val)


@pytest.mark.parametrize("val", [1.2, "1", "characters|not|allowed", ""])
def test_is_link_value_wrong(val):
    assert not is_link_value(val)


@pytest.mark.parametrize("val", ["_1", "abc_2"])
def test_is_valid_resource_id_correct(val):
    assert is_valid_resource_id(val)


@pytest.mark.parametrize("val", [1.2, "1", "characters|not|allowed", " ", " not_good", None])
def test_is_valid_resource_id_wrong(val):
    assert not is_valid_resource_id(val)


@pytest.mark.parametrize(
    "val",
    [
        "2019-10-23T13:45:12.01-14:00",
        "2019-10-23T13:45:12-14:00",
        "2019-10-23T13:45:12Z",
        "2019-10-23T13:45:12-13:30",
        "2019-10-23T13:45:12+01:00",
        "2019-10-23T13:45:12.1111111+01:00",
        "2019-10-23T13:45:12.123456789012Z",
    ],
)
def test_is_timestamp_correct(val: Any) -> None:
    assert is_timestamp(val)


@pytest.mark.parametrize(
    "val",
    [
        True,
        10.0,
        5,
        "2019-10-2",
        "CE:1849:CE:1850",
        "2019-10-23T13:45:12.1234567890123Z",
        "2022",
        "GREGORIAN:CE:2014-01-31",
    ],
)
def test_is_timestamp_wrong(val: Any) -> None:
    assert not is_timestamp(val)


@pytest.mark.parametrize(
    "val", ["http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA", "http://rdfh.ch/4A23/DiAmYQzQSzC7cdTo6OJMYA"]
)
def test_is_dsp_iri_correct(val: Any) -> None:
    assert is_dsp_iri(val)


@pytest.mark.parametrize("val", ["http://www.example.org/prefix1/", "ark:/72163/4123-43xc6ivb931-a.2022829"])
def test_is_dsp_iri_wrong(val: Any) -> None:
    assert not is_dsp_iri(val)


@pytest.mark.parametrize("val", ["ark:/72163/4123-43xc6ivb931-a.2022829"])
def test_is_dsp_ark_correct(val: Any) -> None:
    assert is_dsp_ark(val)


@pytest.mark.parametrize("val", ["http://www.example.org/prefix1/", "http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA"])
def test_is_dsp_ark_wrong(val: Any) -> None:
    assert not is_dsp_ark(val)


@pytest.mark.parametrize(
    "val",
    [f"Start Text<{tag}> middle text </{tag}> ending." for tag in KNOWN_XML_TAGS],
)
def test_check_richtext_syntax_correct(val: Any) -> None:
    with warnings.catch_warnings(record=True) as caught_warnings:
        check_richtext_syntax(val)
    assert len(caught_warnings) == 0


@pytest.mark.parametrize(
    "val",
    [
        'Start <a class="salsah-link" href="IRI:test_thing_0:IRI">test_thing_0</a> ending.',
        '<footnote content="footnote text"/>',
    ],
)
def test_check_richtext_syntax_correct_special_cases(val: Any) -> None:
    with warnings.catch_warnings(record=True) as caught_warnings:
        check_richtext_syntax(val)
    assert len(caught_warnings) == 0


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="method-assign,no-untyped-def"

from uuid import uuid4

import pytest
from rdflib import Graph
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.make_rdf_graph.jsonld_utils import serialise_jsonld_for_resource
from dsp_tools.commands.xmlupload.make_rdf_graph.jsonld_utils import serialise_jsonld_for_value
from dsp_tools.commands.xmlupload.make_rdf_graph.make_resource_and_values import create_resource_with_values
from dsp_tools.commands.xmlupload.models.lookup_models import IRILookups
from dsp_tools.commands.xmlupload.models.processed.res import ProcessedResource
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedBoolean
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedLink
from dsp_tools.commands.xmlupload.stash.stash_models import LinkValueStashItem
from dsp_tools.commands.xmlupload.stash.upload_stashed_resptr_props import _make_link_value_create_graph

ONTO_STR = "http://0.0.0.0:3333/ontology/9999/onto/v2#"

ONTO = Namespace(ONTO_STR)
RES_IRI_STR = "http://rdfh.ch/9999/res_one"
TARGET_IRI_STR = "http://rdfh.ch/9999/target_resource"
RES_IRI = URIRef(RES_IRI_STR)
RES_TYPE = ONTO.Resource


@pytest.fixture
def resource_graph() -> Graph:
    lookups = IRILookups(
        project_iri=URIRef("http://rdfh.ch/9999/project"),
        id_to_iri=IriResolver({"res_one": "http://rdfh.ch/9999/res_one"}),
    )
    res = ProcessedResource(
        res_id="resource_id",
        type_iri="http://0.0.0.0:3333/ontology/9999/onto/v2#TestResource",
        label="Special Characters: äöüéèà",
        permissions=None,
        values=[ProcessedBoolean(True, "http://0.0.0.0:3333/ontology/9999/onto/v2#isTrueOrFalse", None, None)],
        file_value=None,
        iiif_uri=None,
        migration_metadata=None,
    )
    return create_resource_with_values(res, None, lookups)


def test_serialise_jsonld_for_resource(resource_graph: Graph) -> None:
    result_json = serialise_jsonld_for_resource(resource_graph)
    expected = {
        "@type": "http://0.0.0.0:3333/ontology/9999/onto/v2#TestResource",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#isTrueOrFalse": {
            "@type": "http://api.knora.org/ontology/knora-api/v2#BooleanValue",
            "http://api.knora.org/ontology/knora-api/v2#booleanValueAsBoolean": {
                "@type": "http://www.w3.org/2001/XMLSchema#boolean",
                "@value": True,
            },
        },
        "http://api.knora.org/ontology/knora-api/v2#attachedToProject": {"@id": "http://rdfh.ch/9999/project"},
        "http://www.w3.org/2000/01/rdf-schema#label": {
            "@type": "http://www.w3.org/2001/XMLSchema#string",
            "@value": "Special Characters: äöüéèà",
        },
    }
    assert result_json == expected


def test_serialise_jsonld_for_value():
    link_stash = LinkValueStashItem(
        res_id=RES_IRI_STR,
        res_type=RES_TYPE,
        value=ProcessedLink("target_resource_id", ONTO.hasLink, None, None, str(uuid4())),
    )
    graph = _make_link_value_create_graph(link_stash, RES_IRI_STR, TARGET_IRI_STR)
    expected = {
        "@id": "http://rdfh.ch/9999/res_one",
        "@type": "http://0.0.0.0:3333/ontology/9999/onto/v2#Resource",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#hasLinkValue": {
            "@type": "http://api.knora.org/ontology/knora-api/v2#LinkValue",
            "http://api.knora.org/ontology/knora-api/v2#linkValueHasTargetIri": {
                "@id": "http://rdfh.ch/9999/target_resource"
            },
        },
    }
    result = serialise_jsonld_for_value(graph, RES_IRI_STR)
    assert result == expected



from dataclasses import dataclass
from typing import Any

import requests
from loguru import logger
from requests import ReadTimeout
from requests import Response

from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.legal_info_client import LegalInfoClient
from dsp_tools.error.exceptions import BadCredentialsError
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.request_utils import RequestParameters
from dsp_tools.utils.request_utils import log_and_raise_timeouts
from dsp_tools.utils.request_utils import log_request
from dsp_tools.utils.request_utils import log_response

TIMEOUT = 60

HTTP_LACKING_PERMISSIONS = 403


@dataclass
class LegalInfoClientLive(LegalInfoClient):
    server: str
    project_shortcode: str
    authentication_client: AuthenticationClient

    def post_copyright_holders(self, copyright_holders: list[str]) -> None:
        """Send a list of new copyright holders to the API"""
        logger.debug(f"POST {len(copyright_holders)} new copyright holders")
        try:
            response = self._post_and_log_request("copyright-holders", copyright_holders)
        except (TimeoutError, ReadTimeout) as err:
            log_and_raise_timeouts(err)
        if response.ok:
            return
        if response.status_code == HTTP_LACKING_PERMISSIONS:
            raise BadCredentialsError(
                "Only a project or system administrator can create new copyright holders. "
                "Your permissions are insufficient for this action."
            )
        else:
            raise BaseError(
                f"An unexpected response with the status code {response.status_code} was received from the API. "
                f"Please consult 'warnings.log' for details."
            )

    def _post_and_log_request(self, endpoint: str, data: list[str]) -> Response:
        url = f"{self.server}/admin/projects/shortcode/{self.project_shortcode}/legal-info/{endpoint}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.authentication_client.get_token()}",
        }
        params = RequestParameters("POST", url, TIMEOUT, {"data": data}, headers)
        log_request(params)
        response = requests.post(
            url=params.url,
            headers=params.headers,
            data=params.data_serialized,
            timeout=params.timeout,
        )
        log_response(response)
        return response

    def get_licenses_of_a_project(self, enabled_only: bool = True) -> list[dict[str, Any]]:
        logger.debug("GET enabled licenses of the project.")
        page_num = 1
        all_data = []
        is_last_page = False
        while not is_last_page:
            response = self._get_one_license_page(page_num, enabled_only)
            response_dict = response.json()
            all_data.extend(response_dict["data"])
            is_last_page = _is_last_page(response_dict)
        return all_data

    def _get_one_license_page(self, page_num: int, enabled_only: bool) -> Response:
        enabled = str(enabled_only).lower()
        url = (
            f"{self.server}/admin/projects/shortcode/{self.project_shortcode}/"
            f"legal-info/licenses?page={page_num}&page-size=25&order=Asc&showOnlyEnabled={enabled}"
        )
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.authentication_client.get_token()}",
        }
        params = RequestParameters(method="GET", url=url, timeout=TIMEOUT, headers=headers)
        log_request(params)
        try:
            response = requests.get(
                url=params.url,
                headers=params.headers,
                timeout=params.timeout,
            )
            log_response(response)
        except (TimeoutError, ReadTimeout) as err:
            log_and_raise_timeouts(err)
        if response.ok:
            return response
        elif response.status_code == HTTP_LACKING_PERMISSIONS:
            raise BadCredentialsError(
                "Only members of a project or system administrators can request the enabled licenses of a project."
                "Your permissions are insufficient for this action."
            )
        else:
            raise BaseError(
                f"An unexpected response with the status code {response.status_code} was received from the API. "
                f"Please consult 'warnings.log' for details."
            )

    def enable_unknown_license(self) -> None:
        escaped_license_iri = "http%3A%2F%2Frdfh.ch%2Flicenses%2Funknown"
        url = (
            f"{self.server}/admin/projects/shortcode/{self.project_shortcode}/"
            f"legal-info/licenses/{escaped_license_iri}/enable"
        )
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.authentication_client.get_token()}",
        }
        params = RequestParameters("POST", url, TIMEOUT, headers=headers)
        log_request(params)
        response = requests.put(
            url=params.url,
            headers=params.headers,
            timeout=params.timeout,
        )
        log_response(response)
        if response.ok:
            pass
        elif response.status_code == HTTP_LACKING_PERMISSIONS:
            raise BadCredentialsError(
                "Only members of a project or system administrators can enable licenses. "
                "Your permissions are insufficient for this action."
            )
        else:
            raise BaseError(
                f"An unexpected response with the status code {response.status_code} was received from the API. "
                f"Please consult 'warnings.log' for details."
            )


def _is_last_page(response: dict[str, Any]) -> bool:
    current_page = response["pagination"]["currentPage"]
    total_page = response["pagination"]["totalPages"]
    return bool(current_page == total_page)



import regex


def is_prod_like_server(server: str) -> bool:
    return server in [
        "https://api.dasch.swiss",
        "https://api.rdu.dasch.swiss",
        "https://api.ls-prod-server.dasch.swiss",
        "https://api.ls-test-server.dasch.swiss",
        "https://api.stage.dasch.swiss",
    ]


def is_uri(s: str) -> bool:
    """Checks if the given string is a valid URI."""
    # URI = scheme ":" ["//" host [":" port]] path ["?" query] ["#" fragment]
    chars_for_path_query_fragment = r"\w_.\-:~%()!@,;/=*&'+"
    scheme = r"(?<scheme>[a-z][a-z0-9+.\-]*)"
    host = r"(?<host>[\w_.\-:~\[\]]+)"
    port = r"(?<port>:\d{0,6})"
    path = rf"(?<path>/[{chars_for_path_query_fragment}]*)"
    query = rf"(?<query>\?[{chars_for_path_query_fragment}]+)"
    fragment = rf"(?<fragment>#[{chars_for_path_query_fragment}]*)"
    m = regex.match(rf"{scheme}:(//{host}{port}?){path}*{query}*{fragment}?", s, flags=regex.UNICODE)
    return m is not None


def is_iiif_uri(uri: str) -> bool:
    """
    Checks if the given URL is a valid IIIF URL.
    It should support all versions of IIIF servers,
    but was constructed with the syntax described in: https://iiif.io/api/image/3.0/#2-uri-syntax.
    IIIF is open to modifications which are not described in the official documentation
    (see https://iiif.io/api/annex/notes/design_principles/#define-success-not-failure).

    Args:
        uri: The URI to be checked.

    Returns:
        True if the URI is a valid IIIF URI, False otherwise.
    """
    # {scheme}://{server}/{prefix}/{identifier}/{region}/{size}/{rotation}/{quality}.{format}
    split_uri = uri.lower().split("/")
    try:
        region_seg, size_seg, rotation_seg, quality_format_seg = split_uri[-4:]
    except ValueError:
        return False
    # %5E is the URL encoded version of ^ ->
    # because we do change the uri to lower case we need to change that in the regex
    # (\d+(\.\d+)?) -> number, can be integer or float
    # everything needs to be encapsulated in a group
    # otherwise the ^ and $ only apply respectively to the first and last listed in the alternative options
    region_re = (
        r"^(full|square|"  # full | square
        r"((pct:)?(\d+(\.\d+)?,){3}(\d+(\.\d+)?)))$"  # x,y,w,h | pct:x,y,w,h
    )
    if not regex.search(region_re, region_seg):
        return False
    size_re = (
        r"^((\^|%5e)?max|"  # max | ^max
        r"(\^|%5e)?full|"  # full | ^full
        r"(\^|%5e)?pct:\d+(\.\d+)?|"  # pct:n | ^pct:n
        r"(\^|%5e)?(\d+(\.\d+)?)+(,|%2c)|"  # w, | ^w,
        r"(\^|%5e)?,\d+(\.\d+)?|"  # ,h | ^,h
        r"(\^|%5e)?!?\d+(\.\d+)?,\d+(\.\d+)?)$"  # w,h | ^w,h | !w,h | ^!w,h
    )
    if not regex.search(size_re, size_seg):
        return False
    # rotation -> floating point number 0-360 -> n | !n (positive and negative are allowed)
    rotation_re = r"^!?[+-]?\d+(\.\d+)?$"
    if not regex.search(rotation_re, rotation_seg):
        return False
    # quality -> color | colour | gray | grey | bitonal | default | native
    # format -> jpg | tif | png | gif | jp2 | pdf | webp
    quality_format_re = r"^(colou?r|gr[ae]y|bitonal|default|native)(\.(jpg|tif|png|jp2|gif|pdf|webp))?$"
    return bool(regex.search(quality_format_re, quality_format_seg))



import json
from typing import Any

import pytest
import regex

from dsp_tools.commands.project.create.project_create_ontologies import _sort_prop_classes
from dsp_tools.commands.project.create.project_create_ontologies import _sort_resources
from dsp_tools.commands.project.create.project_validate import _check_for_duplicate_res_and_props
from dsp_tools.commands.project.create.project_validate import _check_for_undefined_cardinalities
from dsp_tools.commands.project.create.project_validate import _check_for_undefined_super_property
from dsp_tools.commands.project.create.project_validate import _check_for_undefined_super_resource
from dsp_tools.commands.project.create.project_validate import _collect_link_properties
from dsp_tools.commands.project.create.project_validate import _identify_problematic_cardinalities
from dsp_tools.commands.project.create.project_validate import validate_project
from dsp_tools.error.exceptions import BaseError
from dsp_tools.error.exceptions import InputError
from dsp_tools.utils.json_parsing import parse_json_input


@pytest.fixture
def tp_systematic() -> dict[str, Any]:
    tp_systematic_file = "testdata/json-project/test-project-systematic.json"
    with open(tp_systematic_file, encoding="utf-8") as json_file:
        tp_systematic: dict[str, Any] = json.load(json_file)
    return tp_systematic


@pytest.fixture
def tp_systematic_ontology(tp_systematic: dict[str, Any]) -> dict[str, Any]:
    onto: dict[str, Any] = tp_systematic["project"]["ontologies"][0]
    return onto


@pytest.fixture
def tp_circular_ontology() -> dict[str, Any]:
    tp_circular_ontology_file = "testdata/invalid-testdata/json-project/circular-ontology.json"
    with open(tp_circular_ontology_file, encoding="utf-8") as json_file:
        tp_circular_ontology: dict[str, Any] = json.load(json_file)
    return tp_circular_ontology


def test_sort_resources(tp_systematic_ontology: dict[str, Any]) -> None:
    onto_name: str = tp_systematic_ontology["name"]
    unsorted_resources: list[dict[str, Any]] = tp_systematic_ontology["resources"]
    sorted_resources = _sort_resources(unsorted_resources, onto_name)

    unsorted_resources = sorted(unsorted_resources, key=lambda a: str(a["name"]))
    sorted_resources = sorted(sorted_resources, key=lambda a: str(a["name"]))

    assert unsorted_resources == sorted_resources


def test_sort_prop_classes(tp_systematic_ontology: dict[str, Any]) -> None:
    onto_name: str = tp_systematic_ontology["name"]
    unsorted_props: list[dict[str, Any]] = tp_systematic_ontology["resources"]
    sorted_props = _sort_prop_classes(unsorted_props, onto_name)

    unsorted_props = sorted(unsorted_props, key=lambda a: str(a["name"]))
    sorted_props = sorted(sorted_props, key=lambda a: str(a["name"]))

    assert unsorted_props == sorted_props


def test_validate_project(tp_systematic: dict[str, Any], tp_circular_ontology: dict[str, Any]) -> None:
    assert validate_project(tp_systematic) is True

    with pytest.raises(BaseError, match=regex.escape("Input 'fantasy.xyz' is neither a file path nor a JSON object.")):
        validate_project("fantasy.xyz")

    with pytest.raises(BaseError, match=regex.escape("validation error: 'hasColor' does not match")):
        validate_project("testdata/invalid-testdata/json-project/invalid-super-property.json")

    with pytest.raises(BaseError, match=regex.escape("Your ontology contains properties derived from 'hasLinkTo'")):
        validate_project(tp_circular_ontology)

    with pytest.raises(InputError, match=regex.escape("Listnode names must be unique across all lists")):
        validate_project("testdata/invalid-testdata/json-project/duplicate-listnames.json")


def test_check_for_undefined_cardinalities(tp_systematic: dict[str, Any]) -> None:
    tp_nonexisting_cardinality_file = "testdata/invalid-testdata/json-project/nonexisting-cardinality.json"
    with open(tp_nonexisting_cardinality_file, encoding="utf-8") as json_file:
        tp_nonexisting_cardinality: dict[str, Any] = json.load(json_file)

    assert _check_for_undefined_cardinalities(tp_systematic) is True

    with pytest.raises(
        BaseError,
        match=r"Your data model contains cardinalities with invalid propnames:\n"
        r" - Ontology 'nonexisting-cardinality-onto', resource 'TestThing': "
        r"\[':CardinalityThatWasNotDefinedInPropertiesSection'\]",
    ):
        _check_for_undefined_cardinalities(tp_nonexisting_cardinality)


def test_check_for_undefined_super_property(tp_systematic: dict[str, Any]) -> None:
    tp_nonexisting_super_property_file = "testdata/invalid-testdata/json-project/nonexisting-super-property.json"
    with open(tp_nonexisting_super_property_file, encoding="utf-8") as json_file:
        tp_nonexisting_super_property: dict[str, Any] = json.load(json_file)

    assert _check_for_undefined_super_property(tp_systematic) is True

    with pytest.raises(
        BaseError,
        match=r"Your data model contains properties that are derived from an invalid super-property:\n"
        r" - Ontology 'nonexisting-super-property-onto', property 'hasSimpleText': "
        r"\[':SuperPropertyThatWasNotDefined'\]",
    ):
        _check_for_undefined_super_property(tp_nonexisting_super_property)


def test_check_for_undefined_super_resource(tp_systematic: dict[str, Any]) -> None:
    tp_nonexisting_super_resource_file = "testdata/invalid-testdata/json-project/nonexisting-super-resource.json"
    with open(tp_nonexisting_super_resource_file, encoding="utf-8") as json_file:
        tp_nonexisting_super_resource: dict[str, Any] = json.load(json_file)

    assert _check_for_undefined_super_resource(tp_systematic) is True

    with pytest.raises(
        BaseError,
        match=r"Your data model contains resources that are derived from an invalid super-resource:\n"
        r" - Ontology 'nonexisting-super-resource-onto', resource 'TestThing2': "
        r"\[':SuperResourceThatWasNotDefined'\]",
    ):
        _check_for_undefined_super_resource(tp_nonexisting_super_resource)


def test_circular_references_in_onto(tp_circular_ontology: dict[str, Any]) -> None:
    link_properties = _collect_link_properties(tp_circular_ontology)
    errors = _identify_problematic_cardinalities(tp_circular_ontology, link_properties)
    expected_errors = [
        ("circular-onto:AnyResource", "circular-onto:linkToTestThing1"),
        ("circular-onto:TestThing3", "circular-onto:linkToResource"),
    ]
    assert sorted(errors) == sorted(expected_errors)


def test_parse_json_input() -> None:
    invalid = [
        ("foo/bar", r"The input must be a path to a JSON file or a parsed JSON object"),
        ("testdata/xml-data/test-data-systematic.xml", r"cannot be parsed to a JSON object"),
    ]
    for inv, err_msg in invalid:
        with pytest.raises(BaseError, match=err_msg):
            parse_json_input(inv)


def test_check_for_duplicate_resources() -> None:
    tp_duplicate_resource_file = "testdata/invalid-testdata/json-project/duplicate-resource.json"
    with open(tp_duplicate_resource_file, encoding="utf-8") as json_file:
        tp_duplicate_resource: dict[str, Any] = json.load(json_file)

    with pytest.raises(
        BaseError,
        match=r"Resource names and property names must be unique inside every ontology\.\n"
        r"Resource 'anotherResource' appears multiple times in the ontology 'testonto'\.\n"
        r"Resource 'minimalResource' appears multiple times in the ontology 'testonto'\.\n",
    ):
        _check_for_duplicate_res_and_props(tp_duplicate_resource)


def test_check_for_duplicate_properties() -> None:
    tp_duplicate_property_file = "testdata/invalid-testdata/json-project/duplicate-property.json"
    with open(tp_duplicate_property_file, encoding="utf-8") as json_file:
        tp_duplicate_property: dict[str, Any] = json.load(json_file)
    with pytest.raises(
        BaseError,
        match=r"Resource names and property names must be unique inside every ontology\.\n"
        r"Property 'hasInt' appears multiple times in the ontology 'testonto'\.\n"
        r"Property 'hasText' appears multiple times in the ontology 'testonto'\.\n",
    ):
        _check_for_duplicate_res_and_props(tp_duplicate_property)


if __name__ == "__main__":
    pytest.main([__file__])
